# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

# Most of this file is copied from the C header file.

import ctypes
import ctypes.util
import os
import platform
import sys
from enum import IntEnum, IntFlag
from typing import Any, Protocol

assert sys.platform == "win32", "unsupported platform"


# Configures the behaviour of ctypes.util.find_library().
if "GALLIA_VXLAPI_PATH" in os.environ:
    gallia_setting = os.environ["GALLIA_VXLAPI_PATH"]
    os.environ["PATH"] = os.path.dirname(gallia_setting) + os.pathsep + os.environ["PATH"]  # noqa: PTH120


DLL_NAME = "vxlapi64" if platform.architecture()[0] == "64bit" else "vxlapi"
if dll_path := ctypes.util.find_library(DLL_NAME):
    _xlapi_dll = ctypes.windll.LoadLibrary(dll_path)
else:
    raise FileNotFoundError(f"Vector XL library not found: {DLL_NAME}")


XLuint64 = ctypes.c_int64
XLaccess = XLuint64
XLhandle = ctypes.c_void_p
XLstatus = ctypes.c_short
XLportHandle = ctypes.c_long
XLeventTag = ctypes.c_ubyte
XLstringType = ctypes.c_char_p

xlGetErrorString = _xlapi_dll.xlGetErrorString
xlGetErrorString.argtypes = [XLstatus]
xlGetErrorString.restype = XLstringType


class VectorError(Exception):
    def __init__(self, error_code: int, error_string: str, function: str):
        msg = f'[VectorError {error_code}] in {function}: "{error_string}"'
        super().__init__(msg)


class VectorInitializationError(VectorError):
    pass


class VectorOperationError(VectorError):
    pass


class VectorQueueIsFullError(VectorError):
    pass


class VectorQueueIsEmptyError(VectorError):
    pass


class XL_Status(IntEnum):
    XL_SUCCESS = 0  # =0x0000
    XL_PENDING = 1  # =0x0001
    XL_ERR_QUEUE_IS_EMPTY = 10  # =0x000A
    XL_ERR_QUEUE_IS_FULL = 11  # =0x000B
    XL_ERR_TX_NOT_POSSIBLE = 12  # =0x000C
    XL_ERR_NO_LICENSE = 14  # =0x000E
    XL_ERR_WRONG_PARAMETER = 101  # =0x0065
    XL_ERR_TWICE_REGISTER = 110  # =0x006E
    XL_ERR_INVALID_CHAN_INDEX = 111  # =0x006F
    XL_ERR_INVALID_ACCESS = 112  # =0x0070
    XL_ERR_PORT_IS_OFFLINE = 113  # =0x0071
    XL_ERR_CHAN_IS_ONLINE = 116  # =0x0074
    XL_ERR_NOT_IMPLEMENTED = 117  # =0x0075
    XL_ERR_INVALID_PORT = 118  # =0x0076
    XL_ERR_HW_NOT_READY = 120  # =0x0078
    XL_ERR_CMD_TIMEOUT = 121  # =0x0079
    XL_ERR_CMD_HANDLING = 122  # = 0x007A
    XL_ERR_HW_NOT_PRESENT = 129  # =0x0081
    XL_ERR_NOTIFY_ALREADY_ACTIVE = 131  # =0x0083
    XL_ERR_INVALID_TAG = 132  # = 0x0084
    XL_ERR_INVALID_RESERVED_FLD = 133  # = 0x0085
    XL_ERR_INVALID_SIZE = 134  # = 0x0086
    XL_ERR_INSUFFICIENT_BUFFER = 135  # = 0x0087
    XL_ERR_ERROR_CRC = 136  # = 0x0088
    XL_ERR_BAD_EXE_FORMAT = 137  # = 0x0089
    XL_ERR_NO_SYSTEM_RESOURCES = 138  # = 0x008A
    XL_ERR_NOT_FOUND = 139  # = 0x008B
    XL_ERR_INVALID_ADDRESS = 140  # = 0x008C
    XL_ERR_REQ_NOT_ACCEP = 141  # = 0x008D
    XL_ERR_INVALID_LEVEL = 142  # = 0x008E
    XL_ERR_NO_DATA_DETECTED = 143  # = 0x008F
    XL_ERR_INTERNAL_ERROR = 144  # = 0x0090
    XL_ERR_UNEXP_NET_ERR = 145  # = 0x0091
    XL_ERR_INVALID_USER_BUFFER = 146  # = 0x0092
    XL_ERR_INVALID_PORT_ACCESS_TYPE = 147  # = 0x0093
    XL_ERR_NO_RESOURCES = 152  # =0x0098
    XL_ERR_WRONG_CHIP_TYPE = 153  # =0x0099
    XL_ERR_WRONG_COMMAND = 154  # =0x009A
    XL_ERR_INVALID_HANDLE = 155  # =0x009B
    XL_ERR_RESERVED_NOT_ZERO = 157  # =0x009D
    XL_ERR_INIT_ACCESS_MISSING = 158  # =0x009E
    XL_ERR_WRONG_VERSION = 160  # = 0x00A0
    XL_ERR_CANNOT_OPEN_DRIVER = 201  # =0x00C9
    XL_ERR_WRONG_BUS_TYPE = 202  # =0x00CA
    XL_ERR_DLL_NOT_FOUND = 203  # =0x00CB
    XL_ERR_INVALID_CHANNEL_MASK = 204  # =0x00CC
    XL_ERR_NOT_SUPPORTED = 205  # =0x00CD
    XL_ERR_CONNECTION_BROKEN = 210  # =0x00D2
    XL_ERR_CONNECTION_CLOSED = 211  # =0x00D3
    XL_ERR_INVALID_STREAM_NAME = 212  # =0x00D4
    XL_ERR_CONNECTION_FAILED = 213  # =0x00D5
    XL_ERR_STREAM_NOT_FOUND = 214  # =0x00D6
    XL_ERR_STREAM_NOT_CONNECTED = 215  # =0x00D7
    XL_ERR_QUEUE_OVERRUN = 216  # =0x00D8
    XL_ERROR = 255  # =0x00FF

    # CAN FD Error Codes
    XL_ERR_INVALID_DLC = 513  # =0x0201
    XL_ERR_INVALID_CANID = 514  # =0x0202
    XL_ERR_INVALID_FDFLAG_MODE20 = 515  # =0x203
    XL_ERR_EDL_RTR = 516  # =0x204
    XL_ERR_EDL_NOT_SET = 517  # =0x205
    XL_ERR_UNKNOWN_FLAG = 518  # =0x206


def check_status_operation(result, function, arguments):  # type: ignore
    if result > 0:
        raise VectorOperationError(
            result,
            xlGetErrorString(result).decode(),
            function.__name__,
        )
    return result


def check_rxtx_operation(result, function, arguments):  # type: ignore
    match result:
        case XL_Status.XL_ERR_QUEUE_IS_FULL:
            raise VectorQueueIsFullError(
                result,
                xlGetErrorString(result).decode(),
                function.__name__,
            )
        case XL_Status.XL_ERR_QUEUE_IS_EMPTY:
            raise VectorQueueIsEmptyError(
                result,
                xlGetErrorString(result).decode(),
                function.__name__,
            )
        case XL_Status.XL_SUCCESS:
            return result
        case _:
            raise VectorOperationError(
                result,
                xlGetErrorString(result).decode(),
                function.__name__,
            )
    return result


def check_status_initialization(result, function, arguments):  # type: ignore
    if result > 0:
        raise VectorInitializationError(
            result,
            xlGetErrorString(result).decode(),
            function.__name__,
        )
    return result


XLfrEventTag = ctypes.c_ushort

# activate - channel flags
XL_ACTIVATE_NONE = 0
XL_ACTIVATE_RESET_CLOCK = 8  # using this flag with time synchronisation protocols supported by Vector Timesync Service is not recommended


# Extended error codes
# Too many PDUs configured or too less system memory free
XL_ERR_PDU_OUT_OF_MEMORY = 0x0104
# No cluster configuration has been sent to the driver but is needed for the command which failed
XL_ERR_FR_CLUSTERCONFIG_MISSING = 0x0105
# Invalid offset and/or repetition value specified
XL_ERR_PDU_OFFSET_REPET_INVALID = 0x0106
# Specified PDU payload size is invalid (e.g. size is too large) Frame-API: size is different than static payload length configured in cluster config
XL_ERR_PDU_PAYLOAD_SIZE_INVALID = 0x0107
# Too many frames specified in parameter
XL_ERR_FR_NBR_FRAMES_OVERFLOW = 0x0109
# Specified slot-ID exceeds biggest possible ID specified by the cluster configuration
XL_ERR_FR_SLOT_ID_INVALID = 0x010B
# Specified slot cannot be used by Coldstart-Controller because it's already in use by the eRay
XL_ERR_FR_SLOT_ALREADY_OCCUPIED_BY_ERAY = 0x010C
# Specified slot cannot be used by eRay because it's already in use by the Coldstart-Controller
XL_ERR_FR_SLOT_ALREADY_OCCUPIED_BY_COLDC = 0x010D
# Specified slot cannot be used because it's already in use by another application
XL_ERR_FR_SLOT_OCCUPIED_BY_OTHER_APP = 0x010E
# Specified slot is not in correct segment. E.g.: A dynamic slot was specified for startup&sync
XL_ERR_FR_SLOT_IN_WRONG_SEGMENT = 0x010F
# The given frame-multiplexing rule (specified by offset and repetition) cannot be done because some of the slots are already in use
XL_ERR_FR_FRAME_CYCLE_MULTIPLEX_ERROR = 0x0110


XL_FR_FRAMEFLAG_REQ_TXACK = 0x0020  # used for Tx events only
XL_FR_FRAMEFLAG_TXACK_SS = (
    XL_FR_FRAMEFLAG_REQ_TXACK  # indicates TxAck of SingleShot; used for TxAck events only
)
XL_FR_FRAMEFLAG_RX_UNEXPECTED = (
    XL_FR_FRAMEFLAG_REQ_TXACK  # indicates unexpected Rx frame; used for Rx events only
)

XL_FR_FRAMEFLAG_NEW_DATA_TX = (
    0x0040  # flag used with TxAcks to indicate first TxAck after data update
)
XL_FR_FRAMEFLAG_DATA_UPDATE_LOST = (
    0x0080  # flag used with TxAcks indicating that data update has been lost
)

XL_FR_FRAMEFLAG_SYNTAX_ERROR = 0x0200
XL_FR_FRAMEFLAG_CONTENT_ERROR = 0x0400
XL_FR_FRAMEFLAG_SLOT_BOUNDARY_VIOLATION = 0x0800
XL_FR_FRAMEFLAG_TX_CONFLICT = 0x1000
XL_FR_FRAMEFLAG_EMPTY_SLOT = 0x2000
# Only used with TxAcks: Frame has been transmitted. If not set after transmission, an error has occurred.
XL_FR_FRAMEFLAG_FRAME_TRANSMITTED = 0x8000

# XL_FR_SPY_FRAME_EV event: frameError value
XL_FR_SPY_FRAMEFLAG_FRAMING_ERROR = 0x01
XL_FR_SPY_FRAMEFLAG_HEADER_CRC_ERROR = 0x02
XL_FR_SPY_FRAMEFLAG_FRAME_CRC_ERROR = 0x04
XL_FR_SPY_FRAMEFLAG_BUS_ERROR = 0x08

# FlexRay event tags
XL_FR_START_CYCLE = 0x0080
XL_FR_RX_FRAME = 0x0081
XL_FR_TX_FRAME = 0x0082
XL_FR_TXACK_FRAME = 0x0083
XL_FR_INVALID_FRAME = 0x0084
XL_FR_WAKEUP = 0x0085
XL_FR_SYMBOL_WINDOW = 0x0086
XL_FR_ERROR = 0x0087
XL_FR_ERROR_POC_MODE = 0x01
XL_FR_ERROR_SYNC_FRAMES_BELOWMIN = 0x02
XL_FR_ERROR_SYNC_FRAMES_OVERLOAD = 0x03
XL_FR_ERROR_CLOCK_CORR_FAILURE = 0x04
XL_FR_ERROR_NIT_FAILURE = 0x05
XL_FR_ERROR_CC_ERROR = 0x06
XL_FR_STATUS = 0x0088
XL_FR_NM_VECTOR = 0x008A
XL_FR_TRANCEIVER_STATUS = 0x008B
XL_FR_SPY_FRAME = 0x008E
XL_FR_SPY_SYMBOL = 0x008F


# FlexRay XL API

XL_FR_MAX_DATA_LENGTH = 254


class XL_BusTypes(IntFlag):
    XL_BUS_TYPE_NONE = 0  # =0x00000000
    XL_BUS_TYPE_CAN = 1  # =0x00000001
    XL_BUS_TYPE_LIN = 2  # =0x00000002
    XL_BUS_TYPE_FLEXRAY = 4  # =0x00000004
    XL_BUS_TYPE_AFDX = 8  # =0x00000008
    XL_BUS_TYPE_MOST = 16  # =0x00000010
    XL_BUS_TYPE_DAIO = 64  # =0x00000040
    XL_BUS_TYPE_J1708 = 256  # =0x00000100
    XL_BUS_TYPE_KLINE = 2048  # =0x00000800
    XL_BUS_TYPE_ETHERNET = 4096  # =0x00001000
    XL_BUS_TYPE_A429 = 8192  # =0x00002000


class XL_HardwareType(IntEnum):
    XL_HWTYPE_NONE = 0
    XL_HWTYPE_VIRTUAL = 1
    XL_HWTYPE_CANCARDX = 2
    XL_HWTYPE_CANAC2PCI = 6
    XL_HWTYPE_CANCARDY = 12
    XL_HWTYPE_CANCARDXL = 15
    XL_HWTYPE_CANCASEXL = 21
    XL_HWTYPE_CANCASEXL_LOG_OBSOLETE = 23
    XL_HWTYPE_CANBOARDXL = 25
    XL_HWTYPE_CANBOARDXL_PXI = 27
    XL_HWTYPE_VN2600 = 29
    XL_HWTYPE_VN2610 = XL_HWTYPE_VN2600
    XL_HWTYPE_VN3300 = 37
    XL_HWTYPE_VN3600 = 39
    XL_HWTYPE_VN7600 = 41
    XL_HWTYPE_CANCARDXLE = 43
    XL_HWTYPE_VN8900 = 45
    XL_HWTYPE_VN8950 = 47
    XL_HWTYPE_VN2640 = 53
    XL_HWTYPE_VN1610 = 55
    XL_HWTYPE_VN1630 = 57
    XL_HWTYPE_VN1640 = 59
    XL_HWTYPE_VN8970 = 61
    XL_HWTYPE_VN1611 = 63
    XL_HWTYPE_VN5240 = 64
    XL_HWTYPE_VN5610 = 65
    XL_HWTYPE_VN5620 = 66
    XL_HWTYPE_VN7570 = 67
    XL_HWTYPE_VN5650 = 68
    XL_HWTYPE_IPCLIENT = 69
    XL_HWTYPE_VN5611 = 70
    XL_HWTYPE_IPSERVER = 71
    XL_HWTYPE_VN5612 = 72
    XL_HWTYPE_VX1121 = 73
    XL_HWTYPE_VN5601 = 74
    XL_HWTYPE_VX1131 = 75
    XL_HWTYPE_VT6204 = 77
    XL_HWTYPE_VN1630_LOG = 79
    XL_HWTYPE_VN7610 = 81
    XL_HWTYPE_VN7572 = 83
    XL_HWTYPE_VN8972 = 85
    XL_HWTYPE_VN0601 = 87
    XL_HWTYPE_VN5640 = 89
    XL_HWTYPE_VX0312 = 91
    XL_HWTYPE_VH6501 = 94
    XL_HWTYPE_VN8800 = 95
    XL_HWTYPE_IPCL8800 = 96
    XL_HWTYPE_IPSRV8800 = 97
    XL_HWTYPE_CSMCAN = 98
    XL_HWTYPE_VN5610A = 101
    XL_HWTYPE_VN7640 = 102
    XL_HWTYPE_VX1135 = 104
    XL_HWTYPE_VN4610 = 105
    XL_HWTYPE_VT6306 = 107
    XL_HWTYPE_VT6104A = 108
    XL_HWTYPE_VN5430 = 109
    XL_HWTYPE_VTSSERVICE = 110
    XL_HWTYPE_VN1530 = 112
    XL_HWTYPE_VN1531 = 113
    XL_HWTYPE_VX1161A = 114
    XL_HWTYPE_VX1161B = 115
    XL_HWTYPE_VGNSS = 116
    XL_HWTYPE_VXLAPINIC = 118
    XL_MAX_HWTYPE = 120


class XL_InterfaceVersion(IntEnum):
    XL_INTERFACE_VERSION_V2 = 2
    XL_INTERFACE_VERSION_V3 = 3
    XL_INTERFACE_VERSION = XL_INTERFACE_VERSION_V3
    XL_INTERFACE_VERSION_V4 = 4


class XL_BusCapabilities(IntFlag):
    XL_BUS_COMPATIBLE_CAN = 1
    XL_BUS_ACTIVE_CAP_CAN = 1 << 16
    XL_BUS_COMPATIBLE_LIN = 2
    XL_BUS_ACTIVE_CAP_LIN = 2 << 16
    XL_BUS_COMPATIBLE_FLEXRAY = 4
    XL_BUS_ACTIVE_CAP_FLEXRAY = 4 << 16
    XL_BUS_COMPATIBLE_MOST = 16
    XL_BUS_ACTIVE_CAP_MOST = 16 << 16
    XL_BUS_COMPATIBLE_DAIO = 64
    XL_BUS_ACTIVE_CAP_DAIO = 64 << 16
    XL_BUS_COMPATIBLE_J1708 = 256
    XL_BUS_ACTIVE_CAP_J1708 = 256 << 16
    XL_BUS_COMPATIBLE_KLINE = 2048
    XL_BUS_ACTIVE_CAP_KLINE = 2048 << 16
    XL_BUS_COMPATIBLE_ETHERNET = 4096
    XL_BUS_ACTIVE_CAP_ETHERNET = 4096 << 16
    XL_BUS_COMPATIBLE_A429 = 8192
    XL_BUS_ACTIVE_CAP_A429 = 8192 << 16


# function structures


# structure for xlFrSetConfiguration


class CtypeLike(Protocol):
    @property
    def _fields_(self) -> list[tuple[str, Any]]: ...


class IntrospectMixin:
    def __repr__(self: CtypeLike) -> str:
        fields = []
        for name, _ in self._fields_:
            fields.append(f"{name}: {getattr(self, name)}")

        fields_str = ", ".join(fields)
        return f"{self.__class__.__name__}: {fields_str}"


class s_xl_bus_params_data_can(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("bitRate", ctypes.c_uint),
        ("sjw", ctypes.c_ubyte),
        ("tseg1", ctypes.c_ubyte),
        ("tseg2", ctypes.c_ubyte),
        ("sam", ctypes.c_ubyte),
        ("outputMode", ctypes.c_ubyte),
        ("reserved", ctypes.c_ubyte * 7),
        ("canOpMode", ctypes.c_ubyte),
    ]


class s_xl_bus_params_data_canfd(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("arbitrationBitRate", ctypes.c_uint),
        ("sjwAbr", ctypes.c_ubyte),
        ("tseg1Abr", ctypes.c_ubyte),
        ("tseg2Abr", ctypes.c_ubyte),
        ("samAbr", ctypes.c_ubyte),
        ("outputMode", ctypes.c_ubyte),
        ("sjwDbr", ctypes.c_ubyte),
        ("tseg1Dbr", ctypes.c_ubyte),
        ("tseg2Dbr", ctypes.c_ubyte),
        ("dataBitRate", ctypes.c_uint),
        ("canOpMode", ctypes.c_ubyte),
    ]


class s_xl_bus_params_data(IntrospectMixin, ctypes.Union):
    _fields_ = [
        ("can", s_xl_bus_params_data_can),
        ("canFD", s_xl_bus_params_data_canfd),
        ("most", ctypes.c_ubyte * 12),
        ("flexray", ctypes.c_ubyte * 12),
        ("ethernet", ctypes.c_ubyte * 12),
        ("a429", ctypes.c_ubyte * 28),
    ]


class XLbusParams(IntrospectMixin, ctypes.Structure):
    _fields_ = [("busType", ctypes.c_uint), ("data", s_xl_bus_params_data)]


class XLchannelConfig(IntrospectMixin, ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("name", ctypes.c_char * 32),
        ("hwType", ctypes.c_ubyte),
        ("hwIndex", ctypes.c_ubyte),
        ("hwChannel", ctypes.c_ubyte),
        ("transceiverType", ctypes.c_ushort),
        ("transceiverState", ctypes.c_ushort),
        ("configError", ctypes.c_ushort),
        ("channelIndex", ctypes.c_ubyte),
        ("channelMask", XLuint64),
        ("channelCapabilities", ctypes.c_uint),
        ("channelBusCapabilities", ctypes.c_uint),
        ("isOnBus", ctypes.c_ubyte),
        ("connectedBusType", ctypes.c_uint),
        ("busParams", XLbusParams),
        ("_doNotUse", ctypes.c_uint),
        ("driverVersion", ctypes.c_uint),
        ("interfaceVersion", ctypes.c_uint),
        ("raw_data", ctypes.c_uint * 10),
        ("serialNumber", ctypes.c_uint),
        ("articleNumber", ctypes.c_uint),
        ("transceiverName", ctypes.c_char * 32),
        ("specialCabFlags", ctypes.c_uint),
        ("dominantTimeout", ctypes.c_uint),
        ("dominantRecessiveDelay", ctypes.c_ubyte),
        ("recessiveDominantDelay", ctypes.c_ubyte),
        ("connectionInfo", ctypes.c_ubyte),
        ("currentlyAvailableTimestamps", ctypes.c_ubyte),
        ("minimalSupplyVoltage", ctypes.c_ushort),
        ("maximalSupplyVoltage", ctypes.c_ushort),
        ("maximalBaudrate", ctypes.c_uint),
        ("fpgaCoreCapabilities", ctypes.c_ubyte),
        ("specialDeviceStatus", ctypes.c_ubyte),
        ("channelBusActiveCapabilities", ctypes.c_ushort),
        ("breakOffset", ctypes.c_ushort),
        ("delimiterOffset", ctypes.c_ushort),
        ("reserved", ctypes.c_uint * 3),
    ]


class XLdriverConfig(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("dllVersion", ctypes.c_uint),
        ("channelCount", ctypes.c_uint),
        ("reserved", ctypes.c_uint * 10),
        ("channel", XLchannelConfig * 64),
    ]


class s_xl_fr_cluster_configuration(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("busGuardianEnable", ctypes.c_uint),
        ("busGuardianTick", ctypes.c_uint),
        ("externalClockCorrectionMode", ctypes.c_uint),
        ("gColdStartAttempts", ctypes.c_uint),
        ("gListenNoise", ctypes.c_uint),
        ("gMacroPerCycle", ctypes.c_uint),
        ("gMaxWithoutClockCorrectionFatal", ctypes.c_uint),
        ("gMaxWithoutClockCorrectionPassive", ctypes.c_uint),
        ("gNetworkManagementVectorLength", ctypes.c_uint),
        ("gNumberOfMinislots", ctypes.c_uint),
        ("gNumberOfStaticSlots", ctypes.c_uint),
        ("gOffsetCorrectionStart", ctypes.c_uint),
        ("gPayloadLengthStatic", ctypes.c_uint),
        ("gSyncNodeMax", ctypes.c_uint),
        ("gdActionPointOffset", ctypes.c_uint),
        ("gdDynamicSlotIdlePhase", ctypes.c_uint),
        ("gdMacrotick", ctypes.c_uint),
        ("gdMinislot", ctypes.c_uint),
        ("gdMiniSlotActionPointOffset", ctypes.c_uint),
        ("gdNIT", ctypes.c_uint),
        ("gdStaticSlot", ctypes.c_uint),
        ("gdSymbolWindow", ctypes.c_uint),
        ("gdTSSTransmitter", ctypes.c_uint),
        ("gdWakeupSymbolRxIdle", ctypes.c_uint),
        ("gdWakeupSymbolRxLow", ctypes.c_uint),
        ("gdWakeupSymbolRxWindow", ctypes.c_uint),
        ("gdWakeupSymbolTxIdle", ctypes.c_uint),
        ("gdWakeupSymbolTxLow", ctypes.c_uint),
        ("pAllowHaltDueToClock", ctypes.c_uint),
        ("pAllowPassiveToActive", ctypes.c_uint),
        ("pChannels", ctypes.c_uint),
        ("pClusterDriftDamping", ctypes.c_uint),
        ("pDecodingCorrection", ctypes.c_uint),
        ("pDelayCompensationA", ctypes.c_uint),
        ("pDelayCompensationB", ctypes.c_uint),
        ("pExternOffsetCorrection", ctypes.c_uint),
        ("pExternRateCorrection", ctypes.c_uint),
        ("pKeySlotUsedForStartup", ctypes.c_uint),
        ("pKeySlotUsedForSync", ctypes.c_uint),
        ("pLatestTx", ctypes.c_uint),
        ("pMacroInitialOffsetA", ctypes.c_uint),
        ("pMacroInitialOffsetB", ctypes.c_uint),
        ("pMaxPayloadLengthDynamic", ctypes.c_uint),
        ("pMicroInitialOffsetA", ctypes.c_uint),
        ("pMicroInitialOffsetB", ctypes.c_uint),
        ("pMicroPerCycle", ctypes.c_uint),
        ("pMicroPerMacroNom", ctypes.c_uint),
        ("pOffsetCorrectionOut", ctypes.c_uint),
        ("pRateCorrectionOut", ctypes.c_uint),
        ("pSamplesPerMicrotick", ctypes.c_uint),
        ("pSingleSlotEnabled", ctypes.c_uint),
        ("pWakeupChannel", ctypes.c_uint),
        ("pWakeupPattern", ctypes.c_uint),
        ("pdAcceptedStartupRange", ctypes.c_uint),
        ("pdListenTimeout", ctypes.c_uint),
        ("pdMaxDrift", ctypes.c_uint),
        ("pdMicrotick", ctypes.c_uint),
        ("gdCASRxLowMax", ctypes.c_uint),
        ("gChannels", ctypes.c_uint),
        ("vExternOffsetControl", ctypes.c_uint),
        ("vExternRateControl", ctypes.c_uint),
        ("pChannelsMTS", ctypes.c_uint),
        (
            "framePresetData",
            ctypes.c_uint,
        ),  # 16-bit value with data for pre-initializing the Flexray payload data words
        ("reserved", ctypes.c_uint * 15),
    ]


XLfrClusterConfig = s_xl_fr_cluster_configuration


# structure and defines for function xlFrGetChannelConfig
class s_xl_fr_channel_config(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("status", ctypes.c_uint),  # XL_FR_CHANNEL_CFG_STATUS_xxx
        ("cfgMode", ctypes.c_uint),  # XL_FR_CHANNEL_CFG_MODE_xxx
        ("reserved", ctypes.c_uint * 6),
        ("xlFrClusterConfig", XLfrClusterConfig),  # same as used in function xlFrSetConfig
    ]


XLfrChannelConfig = s_xl_fr_channel_config

# defines for XLfrChannelConfig::status and XLbusParams::data::flexray::status
XL_FR_CHANNEL_CFG_STATUS_INIT_APP_PRESENT = 0x01
XL_FR_CHANNEL_CFG_STATUS_CHANNEL_ACTIVATED = 0x02
XL_FR_CHANNEL_CFG_STATUS_VALID_CLUSTER_CFG = 0x04
XL_FR_CHANNEL_CFG_STATUS_VALID_CFG_MODE = 0x08

# defines for XLfrChannelConfig::cfgMode and XLbusParams::data::flexray::cfgMode
XL_FR_CHANNEL_CFG_MODE_SYNCHRONOUS = 1
XL_FR_CHANNEL_CFG_MODE_COMBINED = 2
XL_FR_CHANNEL_CFG_MODE_ASYNCHRONOUS = 3


# defines for xlFrSetMode (frModes)
XL_FR_MODE_NORMAL = 0x00  # setup the VN3000 (eRay) normal operation mode. (default mode)
XL_FR_MODE_COLD_NORMAL = 0x04  # setup the VN3000 (Fujitsu) normal operation mode. (default mode)

# defines for xlFrSetMode (frStartupAttributes)
XL_FR_MODE_NONE = 0x00  # for normal use
XL_FR_MODE_WAKEUP = 0x01  # for wakeup
XL_FR_MODE_COLDSTART_LEADING = 0x02  # Coldstart path initiating the schedule synchronization
XL_FR_MODE_COLDSTART_FOLLOWING = 0x03  # Coldstart path joining other coldstart nodes
XL_FR_MODE_WAKEUP_AND_COLDSTART_LEADING = (
    0x04  # Send Wakeup and Coldstart path initiating the schedule synchronization
)
XL_FR_MODE_WAKEUP_AND_COLDSTART_FOLLOWING = (
    0x05  # Send Wakeup and Coldstart path joining other coldstart nodes
)


class s_xl_fr_set_modes(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("frMode", ctypes.c_uint),
        ("frStartupAttributes", ctypes.c_uint),
        ("reserved", ctypes.c_uint * 30),
    ]


XLfrMode = s_xl_fr_set_modes

# defines for xlFrSetupSymbolWindow
XL_FR_SYMBOL_MTS = 0x01  # defines a MTS (Media Access Test Symbol)
XL_FR_SYMBOL_CAS = 0x02  # defines a CAS (Collision Avoidance Symbol)


# FR transceiver xlFrSetTransceiverMode modes
XL_FR_TRANSCEIVER_MODE_SLEEP = 0x01
XL_FR_TRANSCEIVER_MODE_NORMAL = 0x02
XL_FR_TRANSCEIVER_MODE_RECEIVE_ONLY = 0x03
XL_FR_TRANSCEIVER_MODE_STANDBY = 0x04

# defines for XL_FR_SYNC_PULSE_EV::triggerSource
# XL_FR_SYNC_PULSE_EXTERNAL    =      XL_SYNC_PULSE_EXTERNAL
# XL_FR_SYNC_PULSE_OUR         =           XL_SYNC_PULSE_OUR
# XL_FR_SYNC_PULSE_OUR_SHARED  =    XL_SYNC_PULSE_OUR_SHARED

# defines for xlFrActivateSpy, mode
XL_FR_SPY_MODE_ASYNCHRONOUS = 0x01
# include <poppack.h>

# include <pshpack8.h>

# defines for xlFrSetAcceptanceFilter
# filterStatus
XL_FR_FILTER_PASS = 0x00000000  # maching frame passes the filter
XL_FR_FILTER_BLOCK = 0x00000001  # maching frame is blocked

# filterTypeMask
XL_FR_FILTER_TYPE_DATA = 0x00000001  # specifies a data frame
XL_FR_FILTER_TYPE_NF = 0x00000002  # specifies a null frame in an used cycle
XL_FR_FILTER_TYPE_FILLUP_NF = 0x00000004  # specifies a null frame in an unused cycle

# filterChannelMask
XL_FR_FILTER_CHANNEL_A = 0x00000001  # specifies FlexRay channel A for the PC
XL_FR_FILTER_CHANNEL_B = 0x00000002  # specifies FlexRay channel B for the PC


class s_xl_fr_acceptance_filter(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("filterStatus", ctypes.c_uint),
        ("filterTypeMask", ctypes.c_uint),
        ("filterFirstSlot", ctypes.c_uint),  # beginning of the slot range
        (
            "filterLastSlot",
            ctypes.c_uint,
        ),  #  end of the slot range (can be the same as filterFirstSlot)
        ("filterChannelMask", ctypes.c_uint),  # channel A, B for PC, channel A, B for COB
    ]


XLfrAcceptanceFilter = s_xl_fr_acceptance_filter


# Flags for the flagsChip parameter
XL_FR_CHANNEL_A = 0x01
XL_FR_CHANNEL_B = 0x02
XL_FR_CHANNEL_AB = XL_FR_CHANNEL_A | XL_FR_CHANNEL_B
XL_FR_CC_COLD_A = 0x04  # second CC channel A to initiate the coldstart
XL_FR_CC_COLD_B = 0x08  # second CC channel B to initiate the coldstart
XL_FR_CC_COLD_AB = XL_FR_CC_COLD_A | XL_FR_CC_COLD_B
XL_FR_SPY_CHANNEL_A = 0x10  # Spy mode flags
XL_FR_SPY_CHANNEL_B = 0x20  # Spy mode flags

XL_FR_QUEUE_OVERFLOW = 0x0100  # driver queue overflow


# T_FLEXRAY_FRAME structure flags / defines
# defines for T_FLEXRAY_FRAME member flags
XL_FR_FRAMEFLAG_STARTUP = 0x0001  # indicates a startup frame
XL_FR_FRAMEFLAG_SYNC = 0x0002  # indicates a sync frame
XL_FR_FRAMEFLAG_NULLFRAME = 0x0004  # indicates a null frame
XL_FR_FRAMEFLAG_PAYLOAD_PREAMBLE = 0x0008  # indicates a present payload preamble bit
XL_FR_FRAMEFLAG_FR_RESERVED = 0x0010  # reserved by Flexray protocol


XL_FR_FRAMEFLAG_REQ_TXACK = 0x0020  # used for Tx events only
XL_FR_FRAMEFLAG_TXACK_SS = (
    XL_FR_FRAMEFLAG_REQ_TXACK  # indicates TxAck of SingleShot; used for TxAck events only
)
XL_FR_FRAMEFLAG_RX_UNEXPECTED = (
    XL_FR_FRAMEFLAG_REQ_TXACK  # indicates unexpected Rx frame; used for Rx events only
)

XL_FR_FRAMEFLAG_NEW_DATA_TX = (
    0x0040  # flag used with TxAcks to indicate first TxAck after data update
)
XL_FR_FRAMEFLAG_DATA_UPDATE_LOST = (
    0x0080  # flag used with TxAcks indicating that data update has been lost
)

XL_FR_FRAMEFLAG_SYNTAX_ERROR = 0x0200
XL_FR_FRAMEFLAG_CONTENT_ERROR = 0x0400
XL_FR_FRAMEFLAG_SLOT_BOUNDARY_VIOLATION = 0x0800
XL_FR_FRAMEFLAG_TX_CONFLICT = 0x1000
XL_FR_FRAMEFLAG_EMPTY_SLOT = 0x2000
XL_FR_FRAMEFLAG_FRAME_TRANSMITTED = 0x8000  # Only used with TxAcks: Frame has been transmitted. If not set after transmission, an error has occurred.

# XL_FR_SPY_FRAME_EV event: frameError value
XL_FR_SPY_FRAMEFLAG_FRAMING_ERROR = 0x01
XL_FR_SPY_FRAMEFLAG_HEADER_CRC_ERROR = 0x02
XL_FR_SPY_FRAMEFLAG_FRAME_CRC_ERROR = 0x04
XL_FR_SPY_FRAMEFLAG_BUS_ERROR = 0x08

# XL_FR_SPY_FRAME_EV event: frameCRC value
XL_FR_SPY_FRAMEFLAG_FRAME_CRC_NEW_LAYOUT = 0x80000000

# XL_FR_SPY_FRAME_EV event: frameFlags value
XL_FR_SPY_FRAMEFLAG_STATIC_FRAME = 0x01

# XL_FR_TX_FRAME event: txMode flags
XL_FR_TX_MODE_CYCLIC = 0x01  # 'normal' cyclic mode
XL_FR_TX_MODE_SINGLE_SHOT = 0x02  # sends only a single shot
XL_FR_TX_MODE_NONE = 0xFF  # switch off TX

# XL_FR_TX_FRAME event: incrementSize values
XL_FR_PAYLOAD_INCREMENT_8BIT = 8
XL_FR_PAYLOAD_INCREMENT_16BIT = 16
XL_FR_PAYLOAD_INCREMENT_32BIT = 32
XL_FR_PAYLOAD_INCREMENT_NONE = 0

# XL_FR_STATUS event: statusType (POC status)
XL_FR_STATUS_DEFAULT_CONFIG = 0x00  # indicates the actual state of the POC in operation control
XL_FR_STATUS_READY = 0x01  # ...
XL_FR_STATUS_NORMAL_ACTIVE = 0x02  # ...
XL_FR_STATUS_NORMAL_PASSIVE = 0x03  # ...
XL_FR_STATUS_HALT = 0x04  # ...
XL_FR_STATUS_MONITOR_MODE = 0x05  # ...
XL_FR_STATUS_CONFIG = 0x0F  # ...

XL_FR_STATUS_WAKEUP_STANDBY = 0x10  # indicates the actual state of the POC in the wakeup path
XL_FR_STATUS_WAKEUP_LISTEN = 0x11  # ...
XL_FR_STATUS_WAKEUP_SEND = 0x12  # ...
XL_FR_STATUS_WAKEUP_DETECT = 0x13  # ...

XL_FR_STATUS_STARTUP_PREPARE = 0x20  # indicates the actual state of the POC in the startup path
XL_FR_STATUS_COLDSTART_LISTEN = 0x21  # ...
XL_FR_STATUS_COLDSTART_COLLISION_RESOLUTION = 0x22  # ...
XL_FR_STATUS_COLDSTART_CONSISTENCY_CHECK = 0x23  # ...
XL_FR_STATUS_COLDSTART_GAP = 0x24  # ...
XL_FR_STATUS_COLDSTART_JOIN = 0x25  # ...
XL_FR_STATUS_INTEGRATION_COLDSTART_CHECK = 0x26  # ...
XL_FR_STATUS_INTEGRATION_LISTEN = 0x27  # ...
XL_FR_STATUS_INTEGRATION_CONSISTENCY_CHECK = 0x28  # ...
XL_FR_STATUS_INITIALIZE_SCHEDULE = 0x29  # ...
XL_FR_STATUS_ABORT_STARTUP = 0x2A  # ...
XL_FR_STATUS_STARTUP_SUCCESS = 0x2B  # ...

# XL_FR_ERROR event: XL_FR_ERROR_POC_MODE, errorMode
XL_FR_ERROR_POC_ACTIVE = 0x00  # Indicates the actual error mode of the POC: active (green)
XL_FR_ERROR_POC_PASSIVE = 0x01  # Indicates the actual error mode of the POC: passive (yellow)
XL_FR_ERROR_POC_COMM_HALT = 0x02  # Indicates the actual error mode of the POC: comm-halt (red)

# XL_FR_ERROR event: XL_FR_ERROR_NIT_FAILURE, flags
XL_FR_ERROR_NIT_SENA = 0x100  # Syntax Error during NIT Channel A
XL_FR_ERROR_NIT_SBNA = 0x200  # Slot Boundary Violation during NIT Channel B
XL_FR_ERROR_NIT_SENB = 0x400  # Syntax Error during NIT Channel A
XL_FR_ERROR_NIT_SBNB = 0x800  # Slot Boundary Violation during NIT Channel B

# XL_FR_ERROR event: XL_FR_ERROR_CLOCK_CORR_FAILURE, flags
XL_FR_ERROR_MISSING_OFFSET_CORRECTION = (
    0x00000001  # Set if no sync frames were received. -> no offset correction possible.
)
XL_FR_ERROR_MAX_OFFSET_CORRECTION_REACHED = (
    0x00000002  # Set if max. offset correction limit is reached.
)
XL_FR_ERROR_MISSING_RATE_CORRECTION = (
    0x00000004  # Set if no even/odd sync frames were received -> no rate correction possible.
)
XL_FR_ERROR_MAX_RATE_CORRECTION_REACHED = (
    0x00000008  # Set if max. rate correction limit is reached.
)

# XL_FR_ERROR event: XL_FR_ERROR_CC_ERROR, erayEir
XL_FR_ERROR_CC_PERR = 0x00000040  # Parity Error, data from MHDS (internal ERay error)
XL_FR_ERROR_CC_IIBA = 0x00000200  # Illegal Input Buffer Access (internal ERay error)
XL_FR_ERROR_CC_IOBA = 0x00000400  # Illegal Output Buffer Access (internal ERay error)
XL_FR_ERROR_CC_MHF = (
    0x00000800  # Message Handler Constraints Flag data from MHDF (internal ERay error)
)
XL_FR_ERROR_CC_EDA = 0x00010000  # Error Detection on channel A, data from ACS
XL_FR_ERROR_CC_LTVA = 0x00020000  # Latest Transmit Violation on channel A
XL_FR_ERROR_CC_TABA = 0x00040000  # Transmit Across Boundary on Channel A
XL_FR_ERROR_CC_EDB = 0x01000000  # Error Detection on channel B, data from ACS
XL_FR_ERROR_CC_LTVB = 0x02000000  # Latest Transmit Violation on channel B
XL_FR_ERROR_CC_TABB = 0x04000000  # Transmit Across Boundary on Channel B

# XL_FR_WAKEUP event: wakeupStatus
XL_FR_WAKEUP_UNDEFINED = 0x00  # No wakeup attempt since CONFIG state was left. (e.g. when a wakeup pattern A|B is received)
XL_FR_WAKEUP_RECEIVED_HEADER = 0x01  # Frame header without coding violation received.
XL_FR_WAKEUP_RECEIVED_WUP = 0x02  # Wakeup pattern on the configured wakeup channel received.
XL_FR_WAKEUP_COLLISION_HEADER = (
    0x03  # Detected collision during wakeup pattern transmission received.
)
XL_FR_WAKEUP_COLLISION_WUP = 0x04  # Collision during wakeup pattern transmission received.
XL_FR_WAKEUP_COLLISION_UNKNOWN = 0x05  # Set when the CC stops wakeup.
XL_FR_WAKEUP_TRANSMITTED = 0x06  # Completed the transmission of the wakeup pattern.
XL_FR_WAKEUP_EXTERNAL_WAKEUP = 0x07  # wakeup comes from external
XL_FR_WAKEUP_WUP_RECEIVED_WITHOUT_WUS_TX = 0x10  # wakeupt pattern received from flexray bus
XL_FR_WAKEUP_RESERVED = 0xFF

# XL_FR_SYMBOL_WINDOW event: flags
XL_FR_SYMBOL_STATUS_SESA = 0x01  # Syntax Error in Symbol Window Channel A
XL_FR_SYMBOL_STATUS_SBSA = 0x02  # Slot Boundary Violation in Symbol Window Channel A
XL_FR_SYMBOL_STATUS_TCSA = 0x04  # Transmission Conflict in Symbol Window Channel A
XL_FR_SYMBOL_STATUS_SESB = 0x08  # Syntax Error in Symbol Window Channel B
XL_FR_SYMBOL_STATUS_SBSB = 0x10  # Slot Boundary Violation in Symbol Window Channel B
XL_FR_SYMBOL_STATUS_TCSB = 0x20  # Transmission Conflict in Symbol Window Channel B
XL_FR_SYMBOL_STATUS_MTSA = 0x40  # MTS received in Symbol Window Channel A
XL_FR_SYMBOL_STATUS_MTSB = 0x80  # MTS received in Symbol Window Channel B


# include <pshpack8.h>

XL_FR_RX_EVENT_HEADER_SIZE = 32
XL_FR_MAX_EVENT_SIZE = 512

# Structures for FlexRay events


class s_xl_fr_start_cycle(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("cycleCount", ctypes.c_uint),
        ("vRateCorrection", ctypes.c_int),
        ("vOffsetCorrection", ctypes.c_int),
        ("vClockCorrectionFailed", ctypes.c_uint),
        ("vAllowPassivToActive", ctypes.c_uint),
        ("reserved", ctypes.c_uint * 3),
    ]


XL_FR_START_CYCLE_EV = s_xl_fr_start_cycle


class s_xl_fr_rx_frame(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("flags", ctypes.c_ushort),
        ("headerCRC", ctypes.c_ushort),
        ("slotID", ctypes.c_ushort),
        ("cycleCount", ctypes.c_uint8),
        ("payloadLength", ctypes.c_uint8),
        ("data", ctypes.c_uint8 * XL_FR_MAX_DATA_LENGTH),
    ]


XL_FR_RX_FRAME_EV = s_xl_fr_rx_frame


class s_xl_fr_tx_frame(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("flags", ctypes.c_ushort),
        ("slotID", ctypes.c_ushort),
        ("offset", ctypes.c_uint8),
        ("repetition", ctypes.c_uint8),
        ("payloadLength", ctypes.c_uint8),
        ("txMode", ctypes.c_uint8),
        ("incrementSize", ctypes.c_uint8),
        ("incrementOffset", ctypes.c_uint8),
        ("reserved0", ctypes.c_uint8),
        ("reserved1", ctypes.c_uint8),
        ("data", ctypes.c_ubyte * XL_FR_MAX_DATA_LENGTH),
    ]


XL_FR_TX_FRAME_EV = s_xl_fr_tx_frame


class s_xl_fr_wakeup(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("cycleCount", ctypes.c_uint8),
        ("wakeupStatus", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8 * 6),
    ]


XL_FR_WAKEUP_EV = s_xl_fr_wakeup


class s_xl_fr_symbol_window(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("symbol", ctypes.c_uint),
        ("flags", ctypes.c_uint),
        ("cycleCount", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8 * 7),
    ]


XL_FR_SYMBOL_WINDOW_EV = s_xl_fr_symbol_window


class s_xl_fr_status(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("statusType", ctypes.c_uint),
        ("reserved", ctypes.c_uint),
    ]


XL_FR_STATUS_EV = s_xl_fr_status


class s_xl_fr_nm_vector(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("nmVector", ctypes.c_uint8 * 12),
        ("cycleCount", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8 * 3),
    ]


XL_FR_NM_VECTOR_EV = s_xl_fr_nm_vector


class s_xl_fr_sync_pulse_ev(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("triggerSource", ctypes.c_uint),
        ("reserved", ctypes.c_uint),
        ("time", XLuint64),
    ]


XL_SYNC_PULSE_EV = s_xl_fr_sync_pulse_ev
XL_FR_SYNC_PULSE_EV = XL_SYNC_PULSE_EV


class s_xl_fr_error_poc_mode(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("errorMode", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8 * 4),
    ]


XL_FR_ERROR_POC_MODE_EV = s_xl_fr_error_poc_mode


class s_xl_fr_error_sync_frames(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("evenSyncFramesA", ctypes.c_short),
        ("oddSyncFramesA", ctypes.c_short),
        ("evenSyncFramesB", ctypes.c_short),
        ("oddSyncFramesB", ctypes.c_short),
        ("reserved", ctypes.c_uint),
    ]


XL_FR_ERROR_SYNC_FRAMES_EV = s_xl_fr_error_sync_frames


class s_xl_fr_error_clock_corr_failure(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("evenSyncFramesA", ctypes.c_short),
        ("oddSyncFramesA", ctypes.c_short),
        ("evenSyncFramesB", ctypes.c_short),
        ("oddSyncFramesB", ctypes.c_short),
        ("flags", ctypes.c_uint),
        ("clockCorrFailedCounter", ctypes.c_uint),
        ("reserved", ctypes.c_uint),
    ]


XL_FR_ERROR_CLOCK_CORR_FAILURE_EV = s_xl_fr_error_clock_corr_failure


class s_xl_fr_error_nit_failure(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("flags", ctypes.c_uint),
        ("reserved", ctypes.c_uint),
    ]


XL_FR_ERROR_NIT_FAILURE_EV = s_xl_fr_error_nit_failure


class s_xl_fr_error_cc_error(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("ccError", ctypes.c_uint),
        ("reserved", ctypes.c_uint),
    ]


XL_FR_ERROR_CC_ERROR_EV = s_xl_fr_error_cc_error


class s_xl_fr_error_info(ctypes.Union):
    _fields_ = [
        ("frPocMode", XL_FR_ERROR_POC_MODE_EV),
        ("frSyncFramesBelowMin", XL_FR_ERROR_SYNC_FRAMES_EV),
        ("frSyncFramesOverload", XL_FR_ERROR_SYNC_FRAMES_EV),
        ("frClockCorrectionFailure", XL_FR_ERROR_CLOCK_CORR_FAILURE_EV),
        ("frNitFailure", XL_FR_ERROR_NIT_FAILURE_EV),
        ("frCCError", XL_FR_ERROR_CC_ERROR_EV),
    ]


class s_xl_fr_error(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("tag", ctypes.c_uint8),
        ("cycleCount", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8 * 6),
        ("errorInfo", s_xl_fr_error_info),
    ]


XL_FR_ERROR_EV = s_xl_fr_error


class s_xl_fr_spy_frame(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("frameLength", ctypes.c_uint),
        ("frameError", ctypes.c_uint8),
        ("tssLength", ctypes.c_uint8),
        ("headerFlags", ctypes.c_ushort),
        ("slotID", ctypes.c_ushort),
        ("headerCRC", ctypes.c_ushort),
        ("payloadLength", ctypes.c_uint8),
        ("cycleCount", ctypes.c_uint8),
        ("frameFlags", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8),
        ("frameCRC", ctypes.c_uint),
        ("data", ctypes.c_uint8 * XL_FR_MAX_DATA_LENGTH),
    ]


XL_FR_SPY_FRAME_EV = s_xl_fr_spy_frame


class s_xl_fr_spy_symbol(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("lowLength", ctypes.c_ushort),
        ("reserved", ctypes.c_ushort),
    ]


XL_FR_SPY_SYMBOL_EV = s_xl_fr_spy_symbol


class s_xl_application_notification(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("notifyReason", ctypes.c_uint),
        ("reserved", ctypes.c_uint * 7),
    ]


XL_APPLICATION_NOTIFICATION_EV = s_xl_application_notification


class s_xl_fr_tag_data(IntrospectMixin, ctypes.Union):
    _fields_ = [
        ("frStartCycle", XL_FR_START_CYCLE_EV),
        ("frRxFrame", XL_FR_RX_FRAME_EV),
        ("frTxFrame", XL_FR_TX_FRAME_EV),
        ("frWakeup", XL_FR_WAKEUP_EV),
        ("frSymbolWindow", XL_FR_SYMBOL_WINDOW_EV),
        ("frError", XL_FR_ERROR_EV),
        ("frStatus", XL_FR_STATUS_EV),
        ("frNmVector", XL_FR_NM_VECTOR_EV),
        ("frSyncPulse", XL_FR_SYNC_PULSE_EV),
        ("frSpyFrame", XL_FR_SPY_FRAME_EV),
        ("frSpySymbol", XL_FR_SPY_SYMBOL_EV),
        ("applicationNotification", XL_APPLICATION_NOTIFICATION_EV),
        ("raw", ctypes.c_uint8 * (XL_FR_MAX_EVENT_SIZE - XL_FR_RX_EVENT_HEADER_SIZE)),
    ]


class s_xl_fr_event(IntrospectMixin, ctypes.Structure):
    _fields_ = [
        ("size", ctypes.c_int),
        ("tag", XLfrEventTag),
        ("channelIndex", ctypes.c_short),
        ("userHandle", ctypes.c_int),
        ("flagsChip", ctypes.c_short),
        ("reserved", ctypes.c_ushort),
        ("timeStamp", XLuint64),
        ("timeStampSync", XLuint64),
        ("tagData", s_xl_fr_tag_data),
    ]


XLfrEvent = s_xl_fr_event


xlOpenDriver = _xlapi_dll.xlOpenDriver
xlOpenDriver.argtypes = []
xlOpenDriver.restype = XLstatus
xlOpenDriver.errcheck = check_status_initialization

xlCloseDriver = _xlapi_dll.xlCloseDriver
xlCloseDriver.argtypes = []
xlCloseDriver.restype = XLstatus
xlCloseDriver.errcheck = check_status_operation

xlGetChannelMask = _xlapi_dll.xlGetChannelMask
xlGetChannelMask.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
xlGetChannelMask.restype = XLaccess

xlOpenPort = _xlapi_dll.xlOpenPort
xlOpenPort.argtypes = [
    ctypes.POINTER(XLportHandle),
    ctypes.c_char_p,
    XLaccess,
    ctypes.POINTER(XLaccess),
    ctypes.c_uint,
    ctypes.c_uint,
    ctypes.c_uint,
]
xlOpenPort.restype = XLstatus
xlOpenPort.errcheck = check_status_initialization

xlClosePort = _xlapi_dll.xlClosePort
xlClosePort.argtypes = [XLportHandle]
xlClosePort.restype = XLstatus
xlClosePort.errcheck = check_status_operation

xlActivateChannel = _xlapi_dll.xlActivateChannel
xlActivateChannel.argtypes = [
    XLportHandle,
    XLaccess,
    ctypes.c_uint,
    ctypes.c_uint,
]
xlActivateChannel.restype = XLstatus
xlActivateChannel.errcheck = check_status_operation

xlDeactivateChannel = _xlapi_dll.xlDeactivateChannel
xlDeactivateChannel.argtypes = [XLportHandle, XLaccess]
xlDeactivateChannel.restype = XLstatus
xlDeactivateChannel.errcheck = check_status_operation

xlFrTransmit = _xlapi_dll.xlFrTransmit
xlFrTransmit.argtypes = [
    XLportHandle,
    XLaccess,
    ctypes.POINTER(XLfrEvent),
]
xlFrTransmit.restype = XLstatus
xlFrTransmit.errcheck = check_rxtx_operation

xlFrReceive = _xlapi_dll.xlFrReceive
xlFrReceive.argtypes = [
    XLportHandle,
    ctypes.POINTER(XLfrEvent),
]
xlFrReceive.restype = XLstatus
xlFrReceive.errcheck = check_rxtx_operation

xlFrSetAcceptanceFilter = _xlapi_dll.xlFrSetAcceptanceFilter
xlFrSetAcceptanceFilter.argtypes = [
    XLportHandle,
    XLaccess,
    ctypes.POINTER(XLfrAcceptanceFilter),
]
xlFrSetAcceptanceFilter.restype = XLstatus
xlFrSetAcceptanceFilter.errcheck = check_status_initialization

xlGetKeymanBoxes = _xlapi_dll.xlGetKeymanBoxes
xlGetKeymanBoxes.argtypes = [
    ctypes.POINTER(ctypes.c_uint),
]
xlGetKeymanBoxes.restype = XLstatus
xlGetKeymanBoxes.errcheck = check_status_initialization

xlSetNotification = _xlapi_dll.xlSetNotification
xlSetNotification.argtypes = [
    XLportHandle,
    ctypes.POINTER(XLhandle),
    ctypes.c_int,
]
xlSetNotification.restype = XLstatus
xlSetNotification.errcheck = check_status_initialization

xlGetDriverConfig = _xlapi_dll.xlGetDriverConfig
xlGetDriverConfig.argtypes = [ctypes.POINTER(XLdriverConfig)]
xlGetDriverConfig.restype = XLstatus
xlGetDriverConfig.errcheck = check_status_operation

xlFlushReceiveQueue = _xlapi_dll.xlFlushReceiveQueue
xlFlushReceiveQueue.argtypes = [XLportHandle]
xlFlushReceiveQueue.restype = XLstatus
xlFlushReceiveQueue.errcheck = check_status_operation

xlFrSetConfiguration = _xlapi_dll.xlFrSetConfiguration
xlFrSetConfiguration.argtypes = [XLportHandle, XLaccess, ctypes.POINTER(XLfrClusterConfig)]
xlFrSetConfiguration.restype = XLstatus
xlFrSetConfiguration.errcheck = check_status_initialization

xlFrGetChannelConfiguration = _xlapi_dll.xlFrGetChannelConfiguration
xlFrGetChannelConfiguration.argtypes = [XLportHandle, XLaccess, ctypes.POINTER(XLfrChannelConfig)]
xlFrGetChannelConfiguration.restype = XLstatus
xlFrGetChannelConfiguration.errcheck = check_status_operation



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0


from gallia.services.uds.core.client import UDSClient, UDSRequestConfig
from gallia.services.uds.core.constants import UDSErrorCodes, UDSIsoServices
from gallia.services.uds.core.service import (
    NegativeResponse,
    PositiveResponse,
    SubFunctionRequest,
    SubFunctionResponse,
    UDSRequest,
    UDSResponse,
)
from gallia.services.uds.ecu import ECU

__all__ = [
    "ECU",
    "NegativeResponse",
    "PositiveResponse",
    "SubFunctionRequest",
    "SubFunctionResponse",
    "UDSClient",
    "UDSErrorCodes",
    "UDSIsoServices",
    "UDSRequest",
    "UDSRequestConfig",
    "UDSResponse",
]



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import atexit
import dataclasses
import datetime
import gzip
import io
import json
import logging
import mmap
import os
import shutil
import socket
import sys
import tempfile
import time
import traceback
from collections.abc import Iterator
from enum import Enum, IntEnum, unique
from logging.handlers import QueueHandler, QueueListener
from pathlib import Path
from queue import Queue
from types import TracebackType
from typing import Any, BinaryIO, Self, TextIO, TypeAlias, cast

import zstandard as zstd

gmt_offset = time.localtime().tm_gmtoff
tz = datetime.timezone(datetime.timedelta(seconds=gmt_offset))


@unique
class ColorMode(Enum):
    """ColorMode is used as an argument to :func:`set_color_mode`."""

    #: Colors are always turned on.
    ALWAYS = "always"
    #: Colors are turned off if the target
    #: stream (e.g. stderr) is not a tty.
    AUTO = "auto"
    #: No colors are used. In other words,
    #: no ANSI escape codes are included.
    NEVER = "never"


def resolve_color_mode(mode: ColorMode, stream: TextIO = sys.stderr) -> bool:
    """Sets the color mode of the console log handler.

    :param mode: The available options are described in :class:`ColorMode`.
    :param stream: Used as a reference for :attr:`ColorMode.AUTO`.
    """
    if sys.platform == "win32":
        return False

    match mode:
        case ColorMode.ALWAYS:
            return True
        case ColorMode.AUTO:
            if os.getenv("NO_COLOR") is not None:
                return False
            else:
                return stream.isatty()
        case ColorMode.NEVER:
            return False


# https://stackoverflow.com/a/35804945
def _add_logging_level(level_name: str, level_num: int) -> None:
    method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError(f"{level_name} already defined in logging module")
    if hasattr(logging, method_name):
        raise AttributeError(f"{method_name} already defined in logging module")
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError(f"{method_name} already defined in logger class")

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def for_level(self, message, *args, **kwargs):  # type: ignore
        if self.isEnabledFor(level_num):
            self._log(
                level_num,
                message,
                args,
                **kwargs,
            )

    def to_root(message, *args, **kwargs):  # type: ignore
        logging.log(level_num, message, *args, **kwargs)  # noqa: LOG015

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, for_level)
    setattr(logging, method_name, to_root)


_add_logging_level("TRACE", 5)
_add_logging_level("NOTICE", 25)


@unique
class Loglevel(IntEnum):
    """A wrapper around the constants exposed by python's
    ``logging`` module. Since gallia adds two additional
    loglevel's (``NOTICE`` and ``TRACE``), this class
    provides a type safe way to access the loglevels.
    The level ``NOTICE`` was added to conform better to
    RFC3164. Subsequently, ``TRACE`` was added to have
    a facility for optional debug messages.
    Loglevel describes python specific values for loglevels
    which are required to integrate with the python ecosystem.
    For generic priority values, see :class:`PenlogPriority`.
    """

    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    NOTICE = logging.NOTICE  # type: ignore
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    TRACE = logging.TRACE  # type: ignore


@unique
class PenlogPriority(IntEnum):
    """PenlogPriority holds the values which are written
    to json log records. These values conform to RFC3164
    with the addition of ``TRACE``. Since Python uses different
    int values for the loglevels, there are two enums in
    gallia describing loglevels. PenlogPriority describes
    generic priority values which are included in json
    log records.
    """

    EMERGENCY = 0
    ALERT = 1
    CRITICAL = 2
    ERROR = 3
    WARNING = 4
    NOTICE = 5
    INFO = 6
    DEBUG = 7
    TRACE = 8

    @classmethod
    def from_str(cls, string: str) -> PenlogPriority:
        """Converts a string to an instance of PenlogPriority.
        ``string`` can be a numeric value (0 to 8 inclusive)
        or a string with a case insensitive name of the level
        (e.g. ``debug``).
        """
        if string.isnumeric():
            return cls(int(string, 0))

        match string.lower():
            case "emergency":
                return cls.EMERGENCY
            case "alert":
                return cls.ALERT
            case "critical":
                return cls.CRITICAL
            case "error":
                return cls.ERROR
            case "warning":
                return cls.WARNING
            case "notice":
                return cls.NOTICE
            case "info":
                return cls.INFO
            case "debug":
                return cls.DEBUG
            case "trace":
                return cls.TRACE
            case _:
                raise ValueError(f"{string} not a valid priority")

    @classmethod
    def from_level(cls, value: int) -> PenlogPriority:
        """Converts an int value (e.g. from python's logging module)
        to an instance of this class.
        """
        match value:
            case Loglevel.TRACE:
                return cls.TRACE
            case Loglevel.DEBUG:
                return cls.DEBUG
            case Loglevel.INFO:
                return cls.INFO
            case Loglevel.NOTICE:
                return cls.NOTICE
            case Loglevel.WARNING:
                return cls.WARNING
            case Loglevel.ERROR:
                return cls.ERROR
            case Loglevel.CRITICAL:
                return cls.CRITICAL
            case _:
                raise ValueError("invalid value")

    def to_level(self) -> Loglevel:
        """Converts an instance of PenlogPriority to :class:`Loglevel`."""
        match self:
            case self.TRACE:
                return Loglevel.TRACE
            case self.DEBUG:
                return Loglevel.DEBUG
            case self.INFO:
                return Loglevel.INFO
            case self.NOTICE:
                return Loglevel.NOTICE
            case self.WARNING:
                return Loglevel.WARNING
            case self.ERROR:
                return Loglevel.ERROR
            case self.CRITICAL:
                return Loglevel.CRITICAL
            case _:
                raise ValueError("invalid value")


def setup_logging(
    level: Loglevel | None = None,
    color_mode: ColorMode = ColorMode.AUTO,
    no_volatile_info: bool = False,
    logger_name: str = "gallia",
) -> None:
    """Enable and configure gallia's logging system.
    If this fuction is not called as early as possible,
    the logging system is in an undefined state und might
    not behave as expected. Always use this function to
    initialize gallia's logging. For instance, ``setup_logging()``
    initializes a QueueHandler to avoid blocking calls during
    logging.

    :param level: The loglevel to enable for the console handler.
                  If this argument is None, the env variable
                  ``GALLIA_LOGLEVEL`` (see :doc:`../env`) is read.
    :param file_level: The loglevel to enable for the file handler.
    :param path: The path to the logfile containing json records.
    :param color_mode: The color mode to use for the console.
    """
    if level is None:
        # FIXME: why is this here and not in config?
        if (raw := os.getenv("GALLIA_LOGLEVEL")) is not None:
            level = PenlogPriority.from_str(raw).to_level()
        else:
            level = Loglevel.DEBUG

    # These are slow and not used by gallia.
    logging.logMultiprocessing = False
    logging.logThreads = False
    logging.logProcesses = False

    logger = logging.getLogger(logger_name)
    # LogLevel cannot be 0 (NOTSET), because only the root logger sends it to its handlers then
    logger.setLevel(1)

    # Clean up potentially existing handlers and create a new async QueueHandler for stderr output
    while len(logger.handlers) > 0:
        logger.handlers[0].close()
        logger.removeHandler(logger.handlers[0])
    colored = resolve_color_mode(color_mode)
    add_stderr_log_handler(logger_name, level, no_volatile_info, colored)


def add_stderr_log_handler(
    logger_name: str,
    level: Loglevel,
    no_volatile_info: bool,
    colored: bool,
) -> None:
    queue: Queue[Any] = Queue()
    logger = logging.getLogger(logger_name)
    logger.addHandler(QueueHandler(queue))

    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(level)
    console_formatter = _ConsoleFormatter()

    console_formatter.colored = colored
    stderr_handler.terminator = ""  # We manually handle the terminator while formatting
    if no_volatile_info is False:
        console_formatter.volatile_info = True

    stderr_handler.setFormatter(console_formatter)

    queue_listener = QueueListener(
        queue,
        *[stderr_handler],
        respect_handler_level=True,
    )
    queue_listener.start()
    atexit.register(queue_listener.stop)


def add_zst_log_handler(
    logger_name: str, filepath: Path, file_log_level: Loglevel
) -> _ZstdFileHandler:
    queue: Queue[Any] = Queue()
    logger = get_logger(logger_name)
    qh = QueueHandler(queue)
    logger.addHandler(qh)

    zstd_handler = _ZstdFileHandler(
        filepath,
        queue_handler=qh,
        level=file_log_level,
    )
    zstd_handler.setLevel(file_log_level)
    zstd_handler.setFormatter(_JSONFormatter())

    queue_listener = QueueListener(
        queue,
        *[zstd_handler],
        respect_handler_level=True,
    )
    queue_listener.start()
    zstd_handler.queue_listener = queue_listener
    return zstd_handler


def remove_zst_log_handler(logger_name: str, handler: _ZstdFileHandler) -> None:
    """This function removes the handler from the specified logger and closes it"""
    logger = get_logger(logger_name)
    # It is important to remove the Handler, otherwise it would still receive log messages
    logger.removeHandler(handler.queue_handler)
    handler.close()


@dataclasses.dataclass
class _PenlogRecordV2:
    module: str
    host: str
    data: str
    datetime: str
    priority: int
    version: int
    tags: list[str] | None = None
    line: str | None = None
    stacktrace: str | None = None
    _python_level_no: int | None = None
    _python_level_name: str | None = None
    _python_func_name: str | None = None


_PenlogRecord: TypeAlias = _PenlogRecordV2


def _colorize_msg(data: str, levelno: int) -> tuple[str, int]:
    if sys.platform == "win32" or not sys.stderr.isatty():
        return data, 0

    out = ""
    match levelno:
        case Loglevel.TRACE:
            style = _Color.GRAY.value
        case Loglevel.DEBUG:
            style = _Color.GRAY.value
        case Loglevel.INFO:
            style = _Color.NOP.value
        case Loglevel.NOTICE:
            style = _Color.BOLD.value
        case Loglevel.WARNING:
            style = _Color.YELLOW.value
        case Loglevel.ERROR:
            style = _Color.RED.value
        case Loglevel.CRITICAL:
            style = _Color.RED.value + _Color.BOLD.value
        case _:
            style = _Color.NOP.value

    out += style
    out += data
    out += _Color.RESET.value

    return out, len(style)


def _format_record(  # noqa: PLR0913
    dt: datetime.datetime,
    name: str,
    data: str,
    levelno: int,
    tags: list[str] | None,
    stacktrace: str | None,
    colored: bool = False,
    volatile_info: bool = False,
) -> str:
    msg = ""
    if volatile_info:
        msg += "\33[2K"
    extra_len = 4
    msg += dt.strftime("%b %d %H:%M:%S.%f")[:-3]
    msg += " "
    msg += name
    if tags is not None and len(tags) > 0:
        msg += f" [{', '.join(tags)}]"
    msg += ": "

    if colored:
        tmp_msg, extra_len_tmp = _colorize_msg(data, levelno)
        msg += tmp_msg
        extra_len += extra_len_tmp
    else:
        msg += data

    if volatile_info and levelno <= Loglevel.INFO:
        terminal_width, _ = shutil.get_terminal_size()
        msg = msg[: terminal_width + extra_len - 1]  # Adapt length to invisible ANSI colors
        msg += _Color.RESET.value
        msg += "\r"
    else:
        msg += "\n"

    if stacktrace is not None:
        msg += "\n"
        msg += stacktrace

    return msg


@dataclasses.dataclass
class PenlogRecord:
    module: str
    host: str
    data: str
    datetime: datetime.datetime
    # FIXME: Enums are slow.
    priority: PenlogPriority
    tags: list[str] | None = None
    colored: bool = False
    line: str | None = None
    stacktrace: str | None = None
    _python_level_no: int | None = None
    _python_level_name: str | None = None
    _python_func_name: str | None = None

    def __str__(self) -> str:
        return _format_record(
            dt=self.datetime,
            name=self.module,
            data=self.data,
            levelno=self._python_level_no
            if self._python_level_no is not None
            else self.priority.to_level(),
            tags=self.tags,
            stacktrace=self.stacktrace,
            colored=self.colored,
        )

    @classmethod
    def parse_priority(cls, data: bytes) -> int | None:
        if not data.startswith(b"<"):
            return None

        prio_str = data[1 : data.index(b">")]
        return int(prio_str)

    @classmethod
    def parse_json(cls, data: bytes) -> Self:
        if data.startswith(b"<"):
            data = data[data.index(b">") + 1 :]

        record = json.loads(data.decode())
        if (v := record["version"]) != 2:
            raise json.JSONDecodeError(f"invalid log record version {v}", data.decode(), 0)

        return cls(
            module=record["module"],
            host=record["host"],
            data=record["data"],
            datetime=datetime.datetime.fromisoformat(record["datetime"]),
            priority=PenlogPriority(record["priority"]),
            tags=record["tags"] if "tags" in record else None,
            line=record["line"] if "line" in record else None,
            stacktrace=record["stacktrace"] if "stacktrace" in record else None,
            _python_level_no=record["_python_level_no"] if "_python_level_no" in record else None,
            _python_level_name=record["_python_level_name"]
            if "_python_level_name" in record
            else None,
            _python_func_name=record["_python_func_name"]
            if "_python_func_name" in record
            else None,
        )

    def to_log_record(self) -> logging.LogRecord:
        level = self.priority.to_level()
        timestamp = self.datetime.timestamp()
        msecs = (timestamp - int(timestamp)) * 1000

        lineno = 0
        pathname = ""
        if (line := self.line) is not None:
            pathname, lineno_str = line.rsplit(":", 1)
            lineno = int(lineno_str)

        return logging.makeLogRecord(
            {
                "name": self.module,
                "priority": self.priority,
                "levelno": level,
                "levelname": logging.getLevelName(level),
                "msg": self.data,
                "pathname": pathname,
                "lineno": lineno,
                "created": timestamp,
                "msecs": msecs,
                "host": self.host,
                "tags": self.tags,
            }
        )


class PenlogReader:
    def __init__(self, path: Path) -> None:
        self.path = path if str(path) != "-" else Path("/dev/stdin")
        self.raw_file = self._prepare_for_mmap(self.path)
        self.file_mmap = mmap.mmap(self.raw_file.fileno(), 0, access=mmap.ACCESS_READ)
        self._current_line = b""
        self._current_record: PenlogRecord | None = None
        self._current_record_index = 0
        self._parsed = False
        self._record_offsets: list[int] = []

    def _test_mmap(self, path: Path) -> bool:
        with path.open("rb") as f:
            try:
                mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
                return True
            except ValueError:
                return False

    def _prepare_for_mmap(self, path: Path) -> BinaryIO:
        if path.is_file() and path.suffix in [".zst", ".gz"]:
            tmpfile = tempfile.TemporaryFile()
            match path.suffix:
                case ".zst":
                    with zstd.open(self.path, "rb") as f:
                        shutil.copyfileobj(cast(BinaryIO, f), tmpfile)
                case ".gz":
                    with gzip.open(self.path, "rb") as f:
                        shutil.copyfileobj(cast(BinaryIO, f), tmpfile)

            tmpfile.flush()
            return cast(BinaryIO, tmpfile)

        if path.is_fifo() or self._test_mmap(path) is False:
            tmpfile = tempfile.TemporaryFile()
            with path.open("rb") as f:
                shutil.copyfileobj(f, tmpfile)
            tmpfile.flush()
            return cast(BinaryIO, tmpfile)

        return self.path.open("rb")

    def _parse_file_structure(self) -> None:
        old_offset = self.file_mmap.tell()

        while True:
            self._record_offsets.append(self.file_mmap.tell())

            line = self.file_mmap.readline()
            if line == b"":
                # The last newline char is not relevant, since
                # no data is following.
                del self._record_offsets[-1]
                break

        self.file_mmap.seek(old_offset)
        self._parsed = True

    def _lookup_offset(self, index: int) -> int:
        if index == 0:
            return 0
        if not self._parsed:
            self._parse_file_structure()
        return self._record_offsets[index]

    @property
    def file_size(self) -> int:
        old_offset = self.file_mmap.tell()
        self.file_mmap.seek(0, io.SEEK_END)
        size = self.file_mmap.tell()
        self.file_mmap.seek(old_offset)
        return size

    @property
    def current_record(self) -> PenlogRecord:
        if self._current_record is not None:
            return self._current_record
        return PenlogRecord.parse_json(self._current_line)

    @property
    def current_priority(self) -> int:
        prio = PenlogRecord.parse_priority(self._current_line)
        if prio is None:
            self._current_record = PenlogRecord.parse_json(self._current_line)
            prio = self._current_record.priority
        return prio

    def seek_to_record(self, n: int) -> None:
        self.file_mmap.seek(self._lookup_offset(n))
        self._current_record_index = n

    def seek_to_current_record(self) -> None:
        self.file_mmap.seek(self._lookup_offset(self._current_record_index))

    def seek_to_previous_record(self) -> None:
        self._current_record_index -= 1
        self.seek_to_record(self._current_record_index)

    def seek_to_next_record(self) -> None:
        self._current_record_index += 1
        self.seek_to_record(self._current_record_index)

    def records(
        self,
        priority: PenlogPriority = PenlogPriority.TRACE,
        offset: int = 0,
        reverse: bool = False,
    ) -> Iterator[PenlogRecord]:
        self.seek_to_record(offset)
        if reverse is False:
            while True:
                if self.readline() == b"":
                    break
                if self.current_priority <= priority:
                    yield self.current_record
        else:
            while True:
                self.readline()
                if self.current_priority <= priority:
                    yield self.current_record
                try:
                    self.seek_to_previous_record()
                except IndexError:
                    break

    def readline(self) -> bytes:
        self._current_record = None
        self._current_line = self.file_mmap.readline()
        return self._current_line

    def close(self) -> None:
        self.file_mmap.close()
        self.raw_file.close()

    def __len__(self) -> int:
        if not self._parsed:
            self._parse_file_structure()
        return len(self._record_offsets)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self.close()


@unique
class _Color(Enum):
    NOP = ""
    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    PURPLE = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    GRAY = "\033[0;38;5;245m"


class _JSONFormatter(logging.Formatter):
    def __init__(self) -> None:
        super().__init__()
        self.hostname = socket.gethostname()

    def format(self, record: logging.LogRecord) -> str:
        tags = record.__dict__["tags"] if "tags" in record.__dict__ else None
        stacktrace = self.formatException(record.exc_info) if record.exc_info else None

        penlog_record = _PenlogRecordV2(
            module=record.name,
            host=self.hostname,
            data=record.getMessage(),
            priority=PenlogPriority.from_level(record.levelno).value,
            datetime=datetime.datetime.fromtimestamp(record.created, tz=tz).isoformat(),
            line=f"{record.pathname}:{record.lineno}",
            stacktrace=stacktrace,
            tags=tags,
            _python_level_no=record.levelno,
            _python_level_name=record.levelname,
            _python_func_name=record.funcName,
            version=2,
        )
        return json.dumps(dataclasses.asdict(penlog_record))


class _ConsoleFormatter(logging.Formatter):
    colored: bool = False
    volatile_info: bool = False

    def format(
        self,
        record: logging.LogRecord,
    ) -> str:
        stacktrace = None

        if record.exc_info:
            exc_type, exc_value, exc_traceback = record.exc_info
            assert exc_type
            assert exc_value
            assert exc_traceback

            stacktrace = "\n"
            stacktrace += "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))

        return _format_record(
            dt=datetime.datetime.fromtimestamp(record.created),
            name=record.name,
            data=record.getMessage(),
            levelno=record.levelno,
            tags=record.__dict__["tags"] if "tags" in record.__dict__ else None,
            stacktrace=stacktrace,
            colored=self.colored,
            volatile_info=self.volatile_info,
        )


class _ZstdFileHandler(logging.Handler):
    def __init__(
        self, path: Path, queue_handler: QueueHandler, level: int | str = logging.NOTSET
    ) -> None:
        super().__init__(level)
        self.file = zstd.open(path, "wb")
        self.queue_handler = queue_handler
        self.queue_listener: QueueListener | None = None

    def close(self) -> None:
        """This function closes the queue handler, the queue listener, and the log file."""
        self.queue_handler.close()
        # There might be no queue_listener or it might already be closed (_thread is None)
        if self.queue_listener is not None and self.queue_listener._thread is not None:
            self.queue_listener.stop()
        self.file.flush()
        self.file.close()

    def emit(self, record: logging.LogRecord) -> None:
        prio = PenlogPriority.from_level(record.levelno).value
        data = f"<{prio}>{self.format(record)}"
        if not data.endswith("\n"):
            data += "\n"
        self.file.write(data.encode())


class Logger(logging.Logger):
    def trace(
        self,
        msg: Any,
        *args: Any,
        exc_info: Any = None,
        stack_info: bool = False,
        extra: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        if self.isEnabledFor(Loglevel.TRACE):
            self._log(
                Loglevel.TRACE,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                **kwargs,
            )

    def notice(
        self,
        msg: Any,
        *args: Any,
        exc_info: Any = None,
        stack_info: bool = False,
        extra: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        if self.isEnabledFor(Loglevel.NOTICE):
            self._log(
                Loglevel.NOTICE,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                **kwargs,
            )

    def result(
        self,
        msg: Any,
        *args: Any,
        exc_info: Any = None,
        stack_info: bool = False,
        extra: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        extra = extra if extra is not None else {}
        extra["tags"] = ["result"]
        if self.isEnabledFor(Loglevel.NOTICE):
            self._log(
                Loglevel.NOTICE,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                **kwargs,
            )


logging.setLoggerClass(Logger)


def get_logger(name: str) -> Logger:
    return cast(Logger, logging.getLogger(name))



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import dataclasses
import json
import os
import os.path
import shutil
import signal
import sys
from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from datetime import UTC, datetime
from enum import Enum, unique
from logging import WARNING
from pathlib import Path
from subprocess import CalledProcessError, run
from typing import Any, Protocol, Self, cast

from pydantic import ConfigDict, field_serializer, model_validator

from gallia import exitcodes
from gallia.command.config import Field, GalliaBaseModel, Idempotent
from gallia.db.handler import DBHandler
from gallia.dumpcap import Dumpcap
from gallia.log import _ZstdFileHandler, add_zst_log_handler, get_logger, remove_zst_log_handler, tz
from gallia.power_supply import PowerSupply
from gallia.power_supply.uri import PowerSupplyURI
from gallia.services.uds.core.exception import UDSException
from gallia.transports import BaseTransport, TargetURI
from gallia.utils import camel_to_snake, get_file_log_level


@unique
class FileNames(Enum):
    PROPERTIES_PRE = "PROPERTIES_PRE.json"
    PROPERTIES_POST = "PROPERTIES_POST.json"
    META = "META.json"
    ENV = "ENV"
    LOGFILE = "log.json.zst"


@unique
class HookVariant(Enum):
    PRE = "pre"
    POST = "post"


@dataclasses.dataclass
class RunMeta:
    command: str
    start_time: str
    end_time: str
    exit_code: int
    config: MutableMapping[str, Any]

    def json(self) -> str:
        return json.dumps(dataclasses.asdict(self))


logger = get_logger(__name__)


if sys.platform.startswith("linux") or sys.platform == "darwin":
    import fcntl

    class Flockable(Protocol):
        @property
        def _lock_file_fd(self) -> int | None: ...

    class FlockMixin:
        def _open_lockfile(self, path: Path) -> int | None:
            if not path.exists():
                path.touch()

            logger.notice("opening lockfile…")
            return os.open(path, os.O_RDONLY)

        async def _aquire_flock(self: Flockable) -> None:
            assert self._lock_file_fd is not None

            try:
                # First do a non blocking flock. If waiting is required,
                # log a message and do a blocking wait afterwards.
                fcntl.flock(self._lock_file_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except BlockingIOError:
                logger.notice("waiting for flock…")
                await asyncio.to_thread(fcntl.flock, self._lock_file_fd, fcntl.LOCK_EX)
            logger.info("Acquired lock. Continuing…")

        def _release_flock(self: Flockable) -> None:
            assert self._lock_file_fd is not None
            fcntl.flock(self._lock_file_fd, fcntl.LOCK_UN)
            os.close(self._lock_file_fd)


if sys.platform == "win32":

    class FlockMixin:
        def _open_lockfile(self, path: Path) -> int | None:
            logger.warn("lockfile in windows is not supported")
            return None

        async def _aquire_flock(self) -> None:
            pass

        def _release_flock(self) -> None:
            pass


class BaseCommandConfig(GalliaBaseModel, cli_group="generic", config_section="gallia"):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    verbose: int = Field(
        0,
        description="Increase verbosity of the console log (0: INFO, 1: DEBUG, 2: TRACE)",
        short="v",
    )
    volatile_info: bool = Field(
        True, description="Overwrite log lines with level info or lower in terminal output"
    )
    trace_log: bool = Field(False, description="set the loglevel of the logfile to TRACE")
    pre_hook: str | None = Field(
        None,
        description="shell script to run before the main entry_point",
        metavar="SCRIPT",
        config_section="gallia.hooks",
    )
    post_hook: str | None = Field(
        None,
        description="shell script to run after the main entry_point",
        metavar="SCRIPT",
        config_section="gallia.hooks",
    )
    hooks: bool = Field(
        True, description="execute pre and post hooks", config_section="gallia.hooks"
    )
    lock_file: Path | None = Field(
        None, description="path to file used for a posix lock", metavar="PATH"
    )
    db: Path | None = Field(None, description="Path to sqlite3 database")
    artifacts_base: Path | None = Field(
        None,
        description="Base directory for artifacts. Required to save artifacts such as logs.",
        metavar="DIR",
        config_section="gallia",
    )


class BaseCommand(FlockMixin, ABC):
    """BaseCommand is the baseclass for all gallia commands.
    This class can be used in standalone scripts via the
    gallia command line interface facility.

    This class needs to be subclassed and all the abstract
    methods need to be implemented. The artifacts_dir is
    generated based on the COMMAND, GROUP, SUBGROUP
    properties (falls back to the class name if all three
    are not set).

    The main entry_point is :meth:`entry_point()`.
    """

    # The config type which is accepted by this class
    # This is used for automatically creating the CLI
    CONFIG_TYPE: type[BaseCommandConfig] = BaseCommandConfig

    #: The string which is shown on the cli with --help.
    SHORT_HELP: str | None = None
    #: The string which is shown at the bottom of --help.
    EPILOG: str | None = None

    #: A list of exception types for which tracebacks are
    #: suppressed at the top level. For these exceptions
    #: a log message with level critical is logged.
    CATCHED_EXCEPTIONS: list[type[Exception]] = []

    log_file_handlers: list[_ZstdFileHandler]

    def __init__(self, config: BaseCommandConfig) -> None:
        self.id = camel_to_snake(self.__class__.__name__)
        self.config = config
        self.artifacts_dir: Path | None = None
        self.run_meta = RunMeta(
            command=f"{type(self).__module__}.{type(self).__name__}",
            start_time=datetime.now(tz).isoformat(),
            exit_code=0,
            end_time="",
            config=json.loads(config.model_dump_json()),
        )
        self._lock_file_fd: int | None = None
        self.db_handler: DBHandler | None = None
        self.log_file_handlers = []

    @abstractmethod
    async def run(self) -> int: ...

    def run_hook(self, variant: HookVariant, exit_code: int | None = None) -> None:
        script = self.config.pre_hook if variant == HookVariant.PRE else self.config.post_hook
        if script is None or script == "":
            return

        hook_id = f"{variant.value}-hook"

        argv = sys.argv[:]
        argv[0] = Path(argv[0]).name
        env = {
            "GALLIA_ARTIFACTS_DIR": str(self.artifacts_dir),
            "GALLIA_HOOK": variant.value,
            "GALLIA_INVOCATION": " ".join(argv),
        } | os.environ

        if variant == HookVariant.POST:
            env["GALLIA_META"] = self.run_meta.json()

        if exit_code is not None:
            env["GALLIA_EXIT_CODE"] = str(exit_code)

        try:
            p = run(script, env=env, text=True, capture_output=True, shell=True, check=True)
            stdout = p.stdout
            stderr = p.stderr
        except CalledProcessError as e:
            logger.warning(f"{variant.value}-hook failed (exit code: {p.returncode})")
            stdout = e.stdout
            stderr = e.stderr

        if stdout:
            logger.info(p.stdout.strip(), extra={"tags": [hook_id, "stdout"]})
        if stderr:
            logger.info(p.stderr.strip(), extra={"tags": [hook_id, "stderr"]})

    async def _db_connect_and_insert_run_meta(self, db_path: Path) -> None:
        self.db_handler = DBHandler(db_path)
        await self.db_handler.connect()

        await self.db_handler.insert_run_meta(
            script=self.run_meta.command,
            config=self.config,
            start_time=datetime.now(UTC).astimezone(),
            path=self.artifacts_dir,
        )

    async def _db_finish_run_meta_and_disconnect(self) -> None:
        if self.db_handler is not None and self.db_handler.connection is not None:
            if self.db_handler.meta is not None:
                try:
                    await self.db_handler.complete_run_meta(
                        datetime.now(UTC).astimezone(), self.run_meta.exit_code, self.artifacts_dir
                    )
                except Exception as e:
                    logger.warning(f"Could not write the run meta to the database: {e!r}")

            try:
                await self.db_handler.disconnect()
            # CancelledError appears only on windows; it is unclear why this happens…
            except Exception as e:
                logger.error(f"Could not close the database connection properly: {e!r}")
            except asyncio.exceptions.CancelledError as e:
                logger.error(f"BUG: {e!r} occured. This only seems to happen on windows")
                logger.error(
                    "If you can reproduce this, open an issue: https://github.com/Fraunhofer-AISEC/gallia"
                )

    def _dump_environment(self, path: Path) -> None:
        environ = cast(dict[str, str], os.environ)
        data = [f"{k}={v}" for k, v in environ.items()]
        path.write_text("\n".join(data) + "\n")

    def _add_latest_link(self, path: Path) -> None:
        dirs = list(path.glob("run-*"))
        dirs.sort(key=lambda x: x.name)

        latest_dir = dirs[-1].relative_to(path)

        symlink = path.joinpath("LATEST")
        symlink.unlink(missing_ok=True)
        try:
            symlink.symlink_to(latest_dir)
        except (OSError, NotImplementedError) as e:
            logger.warn(f"symlink error: {e}")

    def prepare_artifacts_dir(self) -> Path | None:
        if self.config.artifacts_base is None:
            logger.notice("Artifacts base folder not defined, no artifacts will be saved!")
            return None

        else:
            command_dir = self.config.artifacts_base.joinpath(self.id)

            _run_dir = f"run-{datetime.now().strftime('%Y%m%d-%H%M%S.%f')}"
            artifacts_dir = command_dir.joinpath(_run_dir).absolute()
            artifacts_dir.mkdir(parents=True)

            self._dump_environment(artifacts_dir.joinpath(FileNames.ENV.value))
            self._add_latest_link(command_dir)

            return artifacts_dir.absolute()

    async def entry_point(self) -> int:
        if (p := self.config.lock_file) is not None:
            try:
                self._lock_file_fd = self._open_lockfile(p)
                await self._aquire_flock()
            except OSError as e:
                logger.critical(f"Unable to lock {p}: {e}")
                return exitcodes.OSFILE

        self.artifacts_dir = self.prepare_artifacts_dir()
        if self.artifacts_dir is not None:
            self.log_file_handlers.append(
                add_zst_log_handler(
                    logger_name="gallia",
                    filepath=self.artifacts_dir.joinpath(FileNames.LOGFILE.value),
                    file_log_level=get_file_log_level(self.config),
                )
            )

        if self.config.hooks:
            self.run_hook(HookVariant.PRE)

        if self.config.db is not None:
            # Explicitly set log level of aiosqlite to WARNING to avoid log spam
            db_logger = get_logger("aiosqlite")
            db_logger.setLevel(WARNING)
            await self._db_connect_and_insert_run_meta(self.config.db)

        exit_code = 0
        try:
            exit_code = await self.run()
        except (KeyboardInterrupt, asyncio.CancelledError) as e:
            logger.debug(f"{self.__class__} got interrupted: {e!r}")
            exit_code = 128 + signal.SIGINT
        # Ensure that META.json gets written in the case a
        # command calls sys.exit().
        except SystemExit as e:
            match e.code:
                case int():
                    exit_code = e.code
                case _:
                    exit_code = exitcodes.SOFTWARE
        except Exception as e:
            for t in self.CATCHED_EXCEPTIONS:
                if isinstance(e, t):
                    # TODO: Map the exitcode to superclass of builtin exceptions.
                    exit_code = exitcodes.IOERR
                    logger.critical(f"Caught expected exception, stack trace on debug level: {e!r}")
                    logger.debug(e, exc_info=True)
                    break
            else:
                exit_code = exitcodes.SOFTWARE
                logger.critical(e, exc_info=True)
        finally:
            self.run_meta.exit_code = exit_code
            self.run_meta.end_time = datetime.now(tz).isoformat()

            if self.db_handler is not None:
                await self._db_finish_run_meta_and_disconnect()

            if self.artifacts_dir is not None:
                self.artifacts_dir.joinpath(FileNames.META.value).write_text(
                    self.run_meta.json() + "\n"
                )
                logger.notice(f"Stored artifacts at {self.artifacts_dir}")

            # Close open log file handlers to ensure logs are properly written
            # to avoid memory leaks and cross-talking log files
            if len(self.log_file_handlers) > 0:
                logger.info("Syncing log files…")
                while len(self.log_file_handlers) > 0:
                    remove_zst_log_handler(
                        logger_name="gallia",
                        handler=self.log_file_handlers.pop(),
                    )

        if self.config.hooks:
            self.run_hook(HookVariant.POST, exit_code)

        if self._lock_file_fd is not None:
            self._release_flock()

        return exit_code


class AsyncScriptConfig(
    BaseCommandConfig,
    ABC,
    cli_group=BaseCommandConfig._cli_group,
    config_section=BaseCommandConfig._config_section,
):
    pass


class AsyncScript(BaseCommand, ABC):
    """AsyncScript is a base class for a asynchronous gallia command.
    To implement an async script, create a subclass and implement
    the .main() method."""

    GROUP = "script"

    async def setup(self) -> None: ...

    @abstractmethod
    async def main(self) -> None: ...

    async def teardown(self) -> None: ...

    async def run(self) -> int:
        await self.setup()
        try:
            await self.main()
        finally:
            await self.teardown()
        return exitcodes.OK


class ScannerConfig(AsyncScriptConfig, cli_group="scanner", config_section="gallia.scanner"):
    dumpcap: bool = Field(
        sys.platform.startswith("linux"), description="Enable/Disable creating a pcap file"
    )
    target: Idempotent[TargetURI] = Field(
        description="URI that describes the target", metavar="TARGET"
    )
    power_supply: Idempotent[PowerSupplyURI] | None = Field(
        None,
        description="URI specifying the location of the relevant opennetzteil server",
        metavar="URI",
    )
    power_cycle: bool = Field(
        False,
        description="use the configured power supply to power-cycle the ECU when needed (e.g. before starting the scan, or to recover bad state during scanning)",
    )
    power_cycle_sleep: float = Field(
        5.0, description="time to sleep after the power-cycle", metavar="SECs"
    )

    @field_serializer("target", "power_supply")
    def serialize_target_uri(self, target_uri: TargetURI | None) -> Any:
        if target_uri is None:
            return None

        return target_uri.raw

    @model_validator(mode="after")
    def check_power_supply_required(self) -> Self:
        if self.power_cycle and self.power_supply is None:
            raise ValueError("power-cycle needs power-supply")

        return self


class Scanner(AsyncScript, ABC):
    """Scanner is a base class for all scanning related commands.
    A scanner has the following properties:

    - It is async.
    - It loads transports via TargetURIs; available via `self.transport`.
    - Controlling PowerSupplies via the opennetzteil API is supported.
    - `setup()` can be overwritten (do not forget to call `super().setup()`)
      for preparation tasks, such as establishing a network connection or
      starting background tasks.
    - pcap logfiles can be recorded via a Dumpcap background task.
    - `teardown()` can be overwritten (do not forget to call `super().teardown()`)
      for cleanup tasks, such as terminating a network connection or background
      tasks.
    - `main()` is the relevant entry_point for the scanner and must be implemented.
    """

    CATCHED_EXCEPTIONS: list[type[Exception]] = [ConnectionError, UDSException]

    def __init__(self, config: ScannerConfig):
        super().__init__(config)
        self.config: ScannerConfig = config
        self.power_supply: PowerSupply | None = None
        self._transport: BaseTransport | None = None
        self.dumpcap: Dumpcap | None = None

    @property
    def transport(self) -> BaseTransport:
        if self._transport is None:
            raise RuntimeError("Transport accessed before first initialization!")
        return self._transport

    @transport.setter
    def transport(self, transport: BaseTransport) -> None:
        if not isinstance(transport, BaseTransport):
            logger.critical(f"Attempting to assign wrong type to transport: {type(transport)}")
        self._transport = transport

    @abstractmethod
    async def main(self) -> None: ...

    async def setup(self) -> None:
        from gallia.plugins.plugin import load_transport

        if self.config.power_supply is not None:
            self.power_supply = await PowerSupply.connect(self.config.power_supply)
            if self.config.power_cycle is True:
                await self.power_supply.power_cycle(
                    self.config.power_cycle_sleep, lambda: asyncio.sleep(2)
                )

        # Start dumpcap as the first subprocess; otherwise network
        # traffic might be missing.
        if self.artifacts_dir and self.config.dumpcap:
            if shutil.which("dumpcap") is None:
                raise RuntimeError("--dumpcap specified but `dumpcap` is not available")
            self.dumpcap = await Dumpcap.start(self.config.target, self.artifacts_dir)
            if self.dumpcap is None:
                logger.error("`dumpcap` could not be started!")
            else:
                await self.dumpcap.sync()

        self.transport = await load_transport(self.config.target).connect(self.config.target)

    async def teardown(self) -> None:
        await self.transport.close()

        if self.dumpcap:
            await self.dumpcap.stop()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
import time
from pathlib import Path

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSErrorCodes, UDSRequestConfig
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class SASeedsDumperConfig(UDSScannerConfig):
    session: AutoInt = Field(
        0x02, description="Set diagnostic session to perform test in", metavar="INT"
    )
    check_session: bool = Field(False, description="Check current session with read DID")
    level: AutoInt = Field(
        0x11, description="Set security access level to request seed from", metavar="INT"
    )
    send_zero_key: int | None = Field(
        None,
        description="Attempt to fool brute force protection by sending an all-zero key after each seed request. The length of the key can be specified or will otherwise be automatically determined.",
        metavar="BYTE_LENGTH",
        const=0,
    )
    determine_key_size_max_length: int = Field(
        1024,
        description="When trying to automatically determine the key size expected by the ECU, test key lengths from 1 up to N bytes.",
        metavar="INT",
    )
    reset: int | None = Field(
        None,
        metavar="N",
        description="Attempt to fool brute force protection by resetting the ECU when needed or after every N-th requested seed.",
        const=0,
    )
    duration: float = Field(
        0,
        description="Run script for N minutes; zero or negative for infinite runtime (default)",
        metavar="FLOAT",
    )
    data_record: HexBytes = Field(
        b"", description="Append an optional data record to each seed request", metavar="HEXSTRING"
    )
    sleep: float | None = Field(
        None,
        description="Attempt to fool brute force protection by sleeping for N seconds between seed requests.",
    )


class SASeedsDumper(UDSScanner):
    """This scanner tries to enable ProgrammingSession and dump seeds for 12h."""

    CONFIG_TYPE = SASeedsDumperConfig
    SHORT_HELP = "dump security access seeds"

    attempt_reset: bool = False
    is_key_length_determined: bool = False
    key_length: int = 0

    def __init__(self, config: SASeedsDumperConfig):
        super().__init__(config)
        self.config: SASeedsDumperConfig = config
        self.implicit_logging = False

    async def request_seed(self, level: int, data: bytes) -> bytes | None:
        resp = await self.ecu.security_access_request_seed(
            level, data, config=UDSRequestConfig(tags=["ANALYZE"])
        )
        if isinstance(resp, NegativeResponse):
            logger.warning(f"ECU replied with an error: {resp}")
            self.attempt_reset = True
            return None
        return resp.security_seed

    # TODO: Optimize by checking well-known lenghts first: 0x60, 125, 250, 512, 1024.
    async def send_key(self, level: int) -> bool:
        """Returns `True` when main loop should exit, e.g. on successful unlock or exhaustion of key length search space."""
        if (
            self.is_key_length_determined is False
            and self.key_length > self.config.determine_key_size_max_length
        ):
            logger.error(
                f"Unable to identify valid key length for SecurityAccess between 1 and {self.config.determine_key_size_max_length} bytes."
            )
            return True

        if not self.is_key_length_determined:
            logger.info(f"No key length given, testing key length {self.key_length}…")

        resp = await self.ecu.security_access_send_key(
            level + 1, bytes(self.key_length), config=UDSRequestConfig(tags=["ANALYZE"])
        )
        if not isinstance(resp, NegativeResponse):
            logger.result(
                f"That's unexpected: Unlocked SA level {g_repr(level)} with all-zero key of length {self.key_length}."
            )
            return True

        if (
            not self.is_key_length_determined
            and resp.response_code == UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
        ):
            logger.debug(
                f"{self.key_length} does not seem to be the correct key length, incrementing."
            )
            self.key_length += 1
        elif not self.is_key_length_determined:
            logger.result(f"The ECU seems to be expecting keys of length {self.key_length}.")
            self.is_key_length_determined = True
        else:
            logger.debug(f"Key was rejected: {resp}")

        return False

    def log_size(self, path: Path, time_delta: float) -> None:
        size = path.stat().st_size / 1024
        size_unit = "KiB"
        rate = size / time_delta * 3600 if time_delta != 0 else 0
        rate_unit = "KiB"
        if rate > 1024:
            rate = rate / 1024
            rate_unit = "MiB"
        if size > 1024:
            size = size / 1024
            size_unit = "MiB"
        logger.notice(f"Dumping seeds with {rate:.2f}{rate_unit}/h: {size:.2f}{size_unit}")

    async def main(self) -> None:
        session = self.config.session
        logger.info(f"scanning in session: {g_repr(session)}")

        resp = await self.ecu.set_session(session)
        if isinstance(resp, NegativeResponse):
            logger.critical(f"could not change to session: {resp}")
            return

        if self.artifacts_dir is None:
            logger.warning("No artifacts base folder given, saving seeds to current working dir!")
            seeds_file = Path.cwd().joinpath("seeds.bin")
        else:
            seeds_file = self.artifacts_dir.joinpath("seeds.bin")
        file = seeds_file.open("wb", buffering=0)
        duration = self.config.duration * 60
        start_time = time.time()
        last_seed = b""
        requests_since_last_reset = 0
        print_speed = False
        if self.config.send_zero_key is None:
            # Set this to True to not suppress log messages or evaluation of
            # replies later on even though we never send a key
            self.is_key_length_determined = True
        elif self.config.send_zero_key > 0:
            # A specific key_length was given, so use it and no automation
            self.is_key_length_determined = True
            self.key_length = self.config.send_zero_key
        else:
            # Start with length 1 in automatic search
            self.key_length = 1

        while duration <= 0 or time.time() - start_time < duration:
            # Print information about current dump speed every `interval` seconds.
            # As request/response times can jitter a few seconds, we 'arm' the print
            # in one half and 'shoot' once in the other half.
            interval = 60
            i = int(time.time() - start_time) % interval
            if i >= interval // 2:
                print_speed = True
            elif i < interval // 2 and print_speed is True:
                self.log_size(seeds_file, time.time() - start_time)
                print_speed = False

            if self.config.reset is not None:
                if (self.config.reset == 0 and self.attempt_reset) or (
                    self.config.reset > 0 and requests_since_last_reset == self.config.reset
                ):
                    logger.info(
                        f"Resetting the ECU after {requests_since_last_reset} seed requests"
                    )
                    await self.ecu.ecu_reset(0x01)

                    if await self.ecu.wait_for_ecu() is False:
                        logger.error("ECU did not respond after reset; exiting…")
                        sys.exit(1)

                    # Re-enter session. Checking/logging will be done a few lines later if required
                    await self.ecu.set_session(session)

                    requests_since_last_reset = 0
                    self.attempt_reset = False

            if self.config.check_session:
                if not await self.ecu.check_and_set_session(self.config.session):
                    logger.error(f"ECU persistently lost session {g_repr(self.config.session)}")
                    sys.exit(1)

            try:
                seed = await self.request_seed(self.config.level, self.config.data_record)
                if seed is None:
                    continue  # Errors are already logged in .request_seed()
            except TimeoutError:
                logger.error("Timeout while requesting seed")
                continue
            except Exception as e:
                logger.critical(f"Error while requesting seed: {g_repr(e)}")
                sys.exit(1)
            finally:
                requests_since_last_reset += 1

            # During detection of key length the same seed might be returned multiple times.
            # This is, however, not considered critical and should not affect statistical evaluations of the seeds.
            if self.is_key_length_determined is False:
                logger.debug("Still trying to find key size, not evaluating/saving seed responses")
            else:
                file.write(seed)

                if last_seed == seed:
                    logger.warning("Received the same seed as before")
                else:
                    logger.info(f"Received seed of length {len(seed)}")

            last_seed = seed

            if self.key_length > 0:
                try:
                    if await self.send_key(self.config.level):
                        break
                except TimeoutError:
                    logger.warning("Timeout while sending key")
                    continue
                except Exception as e:
                    logger.critical(f"Error while sending key: {g_repr(e)}")
                    sys.exit(1)

            if self.config.sleep is not None:
                logger.info(f"Sleeping for {self.config.sleep} seconds between seed requests…")
                await asyncio.sleep(self.config.sleep)

        file.close()
        self.log_size(seeds_file, time.time() - start_time)
        await self.ecu.leave_session(session, sleep=self.config.power_cycle_sleep)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import os
import signal
import sys
from itertools import islice
from pathlib import Path
from typing import cast

from gallia import exitcodes
from gallia.log import ColorMode, PenlogPriority, PenlogReader, resolve_color_mode


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("FILE", nargs="+", type=Path)
    parser.add_argument(
        "-p",
        "--priority",
        metavar="PRIO",
        type=PenlogPriority.from_str,
        default=PenlogPriority.INFO,
        help="maximal message priority",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-t",
        "--tail",
        action="store_true",
        help="jump to tail while parsing max. -n/--lines lines",
    )
    group.add_argument(
        "--head",
        action="store_true",
        help="only print first -n/--lines lines",
    )
    group.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        help="print the log records in reverse order",
    )
    parser.add_argument(
        "-n",
        "--lines",
        type=int,
        default=100,
        help="print the last n lines",
    )
    parser.add_argument(
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="when to use terminal colors",
    )
    return parser.parse_args()


def _main() -> int:
    args = parse_args()

    for file in args.FILE:
        path = cast(Path, file)
        if not (path.is_file() or path.is_fifo() or str(path) == "-"):
            print(f"not a regular file: {path}", file=sys.stderr)
            return 1

        colored = resolve_color_mode(ColorMode(args.color), stream=sys.stdout)

        with PenlogReader(path) as reader:
            record_generator = reader.records(args.priority, reverse=args.reverse)
            if args.head:
                record_generator = islice(record_generator, args.lines)
            elif args.tail:
                record_generator = reader.records(args.priority, offset=-args.lines)

            for record in record_generator:
                record.colored = colored
                print(record, end="")

    return 0


def main() -> None:
    try:
        sys.exit(_main())
    except json.JSONDecodeError as e:
        print(f"invalid file format: {e}", file=sys.stderr)
        sys.exit(exitcodes.DATAERR)
    # BrokenPipeError appears when stuff is piped to | head.
    # This is not an error for hr.
    except BrokenPipeError:
        # https://docs.python.org/3/library/signal.html#note-on-sigpipe
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown.
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(exitcodes.OK)
    except KeyboardInterrupt:
        sys.exit(128 + signal.SIGINT)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from typing import Self

from pydantic import model_validator

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSResponse

logger = get_logger(__name__)


class WriteByIdentifierPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    data_identifier: AutoInt = Field(description="The data identifier", positional=True)
    session: AutoInt = Field(0x01, description="set session perform test in")
    data: HexBytes | None = Field(None, description="The data which should be written")
    data_file: Path | None = Field(
        None, description="The path to a file with the binary data which should be written"
    )

    @model_validator(mode="after")
    def check_data_source(self) -> Self:
        if not (self.data is None) ^ (self.data_file is None):
            raise ValueError("Exactly one of data or data-file is required")

        return self


class WriteByIdentifierPrimitive(UDSScanner):
    """A simple scanner to talk to the write by identifier service"""

    CONFIG_TYPE = WriteByIdentifierPrimitiveConfig
    SHORT_HELP = "WriteDataByIdentifier"

    def __init__(self, config: WriteByIdentifierPrimitiveConfig):
        super().__init__(config)
        self.config: WriteByIdentifierPrimitiveConfig = config

    async def main(self) -> None:
        try:
            if self.config.session != 0x01:
                resp: UDSResponse = await self.ecu.set_session(self.config.session)
                if isinstance(resp, NegativeResponse):
                    logger.critical(f"could not change to session: {resp}")
                    sys.exit(1)
        except Exception as e:
            logger.critical(f"fatal error: {e!r}")
            sys.exit(1)

        if self.config.data is not None:
            data = self.config.data
        else:
            assert self.config.data_file is not None

            with self.config.data_file.open("rb") as file:
                data = file.read()

        resp = await self.ecu.write_data_by_identifier(self.config.data_identifier, data)
        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            logger.result("Success")



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import errno
import socket as s
import struct
import sys
from typing import Self

assert sys.platform == "linux", "unsupported platform"

from pydantic import BaseModel, field_validator

from gallia.log import get_logger
from gallia.transports._can_constants import (
    CAN_ISOTP_EXTEND_ADDR,
    CAN_ISOTP_LL_OPTS,
    CAN_ISOTP_OPTS,
    CAN_ISOTP_RECV_FC,
    CAN_ISOTP_RX_EXT_ADDR,
    CAN_ISOTP_RX_PADDING,
    CAN_ISOTP_TX_PADDING,
    SOL_CAN_ISOTP,
)
from gallia.transports.base import BaseTransport, TargetURI
from gallia.utils import auto_int

logger = get_logger(__name__)


class ISOTPConfig(BaseModel):
    src_addr: int
    dst_addr: int
    is_extended: bool = False
    is_fd: bool = False
    frame_txtime: int = 10
    ext_address: int | None = None
    rx_ext_address: int | None = None
    tx_padding: int | None = None
    rx_padding: int | None = None
    tx_dl: int = 64

    @field_validator(
        "src_addr",
        "dst_addr",
        "ext_address",
        "rx_ext_address",
        "tx_padding",
        "rx_padding",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


class ISOTPTransport(BaseTransport, scheme="isotp"):
    def __init__(self, target: TargetURI, config: ISOTPConfig, sock: s.socket) -> None:
        super().__init__(target)
        self._sock = sock
        self.config = config

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = target if isinstance(target, TargetURI) else TargetURI(target)
        cls.check_scheme(t)

        if t.hostname is None:
            raise ValueError("empty interface")

        config = ISOTPConfig.model_validate(t.qs_flat)
        sock = s.socket(s.PF_CAN, s.SOCK_DGRAM, s.CAN_ISOTP)
        sock.setblocking(False)

        src_addr = cls._calc_flags(config.src_addr, config.is_extended)
        dst_addr = cls._calc_flags(config.dst_addr, config.is_extended)

        cls._setsockopts(
            sock,
            frame_txtime=config.frame_txtime,
            ext_address=config.ext_address,
            rx_ext_address=config.rx_ext_address,
            tx_padding=config.tx_padding,
            rx_padding=config.rx_padding,
        )
        # If CAN-FD is used, jumbo frames are possible.
        # This fails for non-fd configurations.
        if config.is_fd:
            cls._setsockllopts(sock, canfd=config.is_fd, tx_dl=config.tx_dl)

        sock.bind((t.hostname, dst_addr, src_addr))

        return cls(t, config, sock)

    @staticmethod
    def _calc_flags(can_id: int, extended: bool = False) -> int:
        if extended:
            return (can_id & s.CAN_EFF_MASK) | s.CAN_EFF_FLAG
        return can_id & s.CAN_SFF_MASK

    @staticmethod
    def _setsockopts(  # noqa: PLR0913
        sock: s.socket,
        frame_txtime: int,
        tx_padding: int | None = None,
        rx_padding: int | None = None,
        ext_address: int | None = None,
        rx_ext_address: int | None = None,
    ) -> None:
        flags = 0
        if ext_address is not None:
            flags |= CAN_ISOTP_EXTEND_ADDR
        else:
            ext_address = 0

        if rx_ext_address is not None:
            flags |= CAN_ISOTP_RX_EXT_ADDR
        else:
            rx_ext_address = 0

        if tx_padding is not None:
            flags |= CAN_ISOTP_TX_PADDING
        else:
            tx_padding = 0

        if rx_padding is not None:
            flags |= CAN_ISOTP_RX_PADDING
        else:
            rx_padding = 0

        data = struct.pack(
            "@IIBBBB",
            flags,
            frame_txtime,
            ext_address,
            tx_padding,
            rx_padding,
            rx_ext_address,
        )
        sock.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_OPTS, data)

    @staticmethod
    def _setsockfcopts(
        sock: s.socket,
        bs: int = 0,
        stmin: int = 0,
        wftmax: int = 0,
    ) -> None:
        data = struct.pack("@BBB", bs, stmin, wftmax)
        sock.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_RECV_FC, data)

    @staticmethod
    def _setsockllopts(sock: s.socket, canfd: bool, tx_dl: int) -> None:
        canmtu = 72 if canfd else 16
        # The flags are set to 0, since the author marks this as obsolete.
        data = struct.pack("@BBB", canmtu, tx_dl, 0)
        sock.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_LL_OPTS, data)

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        t = tags + ["write"] if tags is not None else ["write"]
        logger.trace(data.hex(), extra={"tags": t})

        loop = asyncio.get_running_loop()
        await asyncio.wait_for(loop.sock_sendall(self._sock, data), timeout)
        return len(data)

    async def read(self, timeout: float | None = None, tags: list[str] | None = None) -> bytes:
        loop = asyncio.get_running_loop()
        try:
            data = await asyncio.wait_for(loop.sock_recv(self._sock, self.BUFSIZE), timeout)
        except OSError as e:
            if e.errno == errno.ECOMM:
                raise BrokenPipeError(f"isotp flow control frame missing: {e}") from e
            if e.errno == errno.EILSEQ:
                raise BrokenPipeError(f"invalid consecutive frame numbers: {e}") from e
            raise e
        logger.trace(data.hex(), extra={"tags": tags})
        return data

    async def close(self) -> None:
        self._sock.close()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import reprlib
import sys
from itertools import product

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, EnumArg, Field, HexBytes, Ranges, Ranges2D
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.client import UDSRequestConfig
from gallia.services.uds.core.constants import RoutineControlSubFuncs, UDSErrorCodes, UDSIsoServices
from gallia.services.uds.core.exception import IllegalResponse
from gallia.services.uds.core.service import NegativeResponse, UDSResponse
from gallia.services.uds.core.utils import g_repr, service_repr
from gallia.services.uds.helpers import suggests_service_not_supported

logger = get_logger(__name__)


class ScanIdentifiersConfig(UDSScannerConfig):
    sessions: Ranges | None = Field(
        None, description="Set list of sessions to be tested; all if None", metavar="SESSION_ID"
    )
    start: AutoInt = Field(0, description="start scan at this dataIdentifier")
    end: AutoInt = Field(0xFFFF, description="end scan at this dataIdentifier")
    payload: HexBytes | None = Field(
        None, description="Payload which will be appended for each request as hex string"
    )
    service: EnumArg[UDSIsoServices] = Field(
        UDSIsoServices.ReadDataByIdentifier,
        description="Service (ID) to scan; defaults to ReadDataByIdentifier;\ncurrently supported:\n0x27 Security Access;\n0x22 Read Data By Identifier;\n0x2e Write Data By Identifier;\n0x31 Routine Control;\n",
    )
    check_session: int | None = Field(
        None,
        description="Check current session via read DID [for every nth DataIdentifier] and try to recover session; only takes affect if --sessions is given",
        const=1,
    )
    skip: Ranges2D = Field(
        {},
        metavar="SESSION_ID:ID",
        description="The data identifiers to be skipped per session.\nA session specific skip is given by <session_id>:<identifiers>\nwhere <identifiers> is a comma separated list of single ids or id ranges using a dash.\nExamples:\n - 0x01:0xf3\n - 0x10-0x2f\n - 0x01:0xf3,0x10-0x2f\nMultiple session specific skips are separated by space.\nOnly takes affect if --sessions is given.\n",
    )
    skip_not_supported: bool = Field(
        False, description="Stop scanning in session if service seems to be not available"
    )


class ScanIdentifiers(UDSScanner):
    """This scanner scans DataIdentifiers of various
    services. Specific requirements such as for RoutineControl or SecurityAccess
    are considered and implemented in the script.
    """

    CONFIG_TYPE = ScanIdentifiersConfig
    SHORT_HELP = "identifier scan of a UDS service"

    def __init__(self, config: ScanIdentifiersConfig):
        super().__init__(config)
        self.config: ScanIdentifiersConfig = config

    async def main(self) -> None:
        if self.config.sessions is None:
            # Attempt to read session once for state update
            try:
                current_session = await self.ecu.read_session(
                    UDSRequestConfig(
                        timeout=2,
                        max_retry=0,
                    )
                )
                logger.notice(f"Performing scan in current session, which is {current_session:#x}")
            except Exception as e:
                logger.notice(f"Performing scan in current (unknown) session: {e!r}")

            if not await self.perform_scan():
                sys.exit(1)
        else:
            sessions: list[int] = [
                s
                for s in self.config.sessions
                if s not in self.config.skip or self.config.skip[s] is not None
            ]
            logger.info(f"testing sessions {g_repr(sessions)}")
            clean_returns = True

            # TODO: Unified shortened output necessary here
            logger.info(f"skipping identifiers {reprlib.repr(self.config.skip)}")

            for session in sessions:
                logger.notice(f"Switching to session {g_repr(session)}")
                resp: UDSResponse = await self.ecu.set_session(session)
                if isinstance(resp, NegativeResponse):
                    logger.warning(f"Switching to session {g_repr(session)} failed: {resp}")
                    continue

                logger.result(f"Starting scan in session: {g_repr(session)}")

                clean_returns = clean_returns and await self.perform_scan(session)

                logger.result(f"Scan in session {g_repr(session)} is complete!")
                logger.info(f"Leaving session {g_repr(session)} via hook")
                await self.ecu.leave_session(session, sleep=self.config.power_cycle_sleep)

            if not clean_returns:
                sys.exit(1)

    async def perform_scan(self, session: None | int = None) -> bool:
        positive_DIDs = 0
        abnormal_DIDs = 0
        timeout_DIDs = 0
        sub_functions = [0x00]

        if self.config.service == UDSIsoServices.RoutineControl:
            if not self.config.payload:
                logger.warning(
                    "Scanning RoutineControl with empty payload can successfully execute some "
                    + "routines that might have irreversible effects without elevated privileges"
                )

            # Scan all three subfunctions (startRoutine, stopRoutine, requestRoutineResults)
            sub_functions = list(map(int, RoutineControlSubFuncs))

        if self.config.service == UDSIsoServices.SecurityAccess and self.config.end > 0x7F:
            logger.warning(
                "Service 0x27 SecurityAccess only accepts subFunctions (7-bit identifiers); "
                + f"limiting END to {g_repr(0x7F)} instead of {g_repr(self.config.end)}"
            )
            self.config.end = 0x7F

        for DID, sub_function in product(
            range(self.config.start, self.config.end + 1), sub_functions
        ):
            if session in self.config.skip and (
                (session_skip := self.config.skip[session]) is None or DID in session_skip
            ):
                logger.info(f"{g_repr(DID)}: skipped")
                continue

            if (
                session is not None
                and self.config.check_session
                and (DID % self.config.check_session == 0)
            ):
                # Check session and try to recover from wrong session (max 3 times), else skip session
                if not await self.ecu.check_and_set_session(session, retries=3):
                    logger.error(
                        f"Aborting scan on session {g_repr(session)}; current DID was {g_repr(DID)}"
                    )
                    return False

            if self.config.service == UDSIsoServices.SecurityAccess:
                pdu = bytes([self.config.service, DID])

            elif self.config.service == UDSIsoServices.RoutineControl:
                pdu = bytes(
                    [self.config.service, sub_function, DID >> 8, DID & 0xFF]
                )  # Needs extra byte for sub function
            else:
                # DefaultBehavior, e.g. for ReadDataByIdentifier/WriteDataByIdentifier
                pdu = bytes([self.config.service, DID >> 8, DID & 0xFF])

            if self.config.payload:
                pdu += self.config.payload

            try:
                resp = await self.ecu.send_raw(
                    pdu, config=UDSRequestConfig(tags=["ANALYZE"], max_retry=3)
                )
            except TimeoutError:
                logger.result(f"{g_repr(DID)}: Retries exceeded")
                timeout_DIDs += 1
                continue
            except IllegalResponse as e:
                logger.warning(g_repr(e))
                continue

            if isinstance(resp, NegativeResponse):
                if suggests_service_not_supported(resp):
                    logger.info(
                        f"{g_repr(DID)}: {resp}; does session {g_repr(session)} support service {service_repr(self.config.service)}?"
                    )

                    if self.config.skip_not_supported:
                        break

                # RequestOutOfRange is a common reply for invalid/unknown DataIdentifiers
                # SubFunctionNotSupported is also not worth to be logged as result
                elif resp.response_code in (
                    UDSErrorCodes.requestOutOfRange,
                    UDSErrorCodes.subFunctionNotSupported,
                ):
                    logger.info(f"{g_repr(DID)}: {resp}")
                else:
                    logger.result(f"{g_repr(DID)}: {resp}")
                    abnormal_DIDs += 1
            else:
                logger.result(f"{g_repr(DID)}: {resp}")
                positive_DIDs += 1

        logger.result(f"Positive replies: {positive_DIDs}")
        logger.result(f"Abnormal replies: {abnormal_DIDs}")
        logger.result(f"Timeouts: {timeout_DIDs}")

        return True



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from typing import Self

from gallia.log import get_logger
from gallia.transports.base import BaseTransport, LinesTransportMixin, TargetURI

logger = get_logger(__name__)


class TCPTransport(BaseTransport, scheme="tcp"):
    def __init__(
        self,
        target: TargetURI,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        super().__init__(target)
        self.reader = reader
        self.writer = writer

    @classmethod
    async def connect(cls, target: str | TargetURI, timeout: float | None = None) -> Self:
        t = target if isinstance(target, TargetURI) else TargetURI(target)
        cls.check_scheme(t)

        reader, writer = await asyncio.wait_for(
            asyncio.open_connection(t.hostname, t.port), timeout
        )
        return cls(t, reader, writer)

    async def close(self) -> None:
        if self.is_closed:
            return
        self.is_closed = True
        self.writer.close()
        await self.writer.wait_closed()

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        t = tags + ["write"] if tags is not None else ["write"]
        logger.trace(data.hex(), extra={"tags": t})

        self.writer.write(data)
        await asyncio.wait_for(self.writer.drain(), timeout)
        return len(data)

    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        data = await asyncio.wait_for(self.reader.read(self.BUFSIZE), timeout)

        t = tags + ["read"] if tags is not None else ["read"]
        logger.trace(data.hex(), extra={"tags": t})
        return data


class TCPLinesTransport(LinesTransportMixin, TCPTransport, scheme="tcp-lines"):
    def get_reader(self) -> asyncio.StreamReader:
        return self.reader

    def get_writer(self) -> asyncio.StreamWriter:
        return self.writer



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import random
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Self

from pydantic import field_serializer, model_validator

from gallia.command import AsyncScript
from gallia.command.base import AsyncScriptConfig
from gallia.command.config import Field, Idempotent
from gallia.log import get_logger
from gallia.services.uds.server import (
    DBUDSServer,
    RandomUDSServer,
    TCPUDSServerTransport,
    UDSServer,
    UDSServerTransport,
)
from gallia.transports import TargetURI, TransportScheme

logger = get_logger(__name__)


class VirtualECUConfig(AsyncScriptConfig):
    target: Idempotent[TargetURI] = Field(positional=True)

    @field_serializer("target")
    def serialize_target_uri(self, target_uri: TargetURI | None) -> Any:
        if target_uri is None:
            return None

        return target_uri.raw

    @model_validator(mode="after")
    def check_transport_requirements(self) -> Self:
        supported: list[TransportScheme] = []

        if sys.platform.startswith("linux"):
            supported = [TransportScheme.TCP, TransportScheme.ISOTP, TransportScheme.UNIX_LINES]

        if sys.platform.startswith("win32"):
            supported = [TransportScheme.TCP]

        if self.target.scheme not in supported:
            raise ValueError(f"Unsupported transport scheme! Use any of {supported}")

        return self


class DbVirtualECUConfig(VirtualECUConfig, DBUDSServer.Behavior):
    path: Path = Field(positional=True)
    ecu: str | None
    properties: dict[str, Any] | None = Field(metavar="PROPERTIES")


class RngVirtualECUConfig(
    VirtualECUConfig, RandomUDSServer.Behavior, RandomUDSServer.RandomnessParameters
):
    seed: int = Field(
        random.randint(0, sys.maxsize),
        description="Set the seed of the internal random number generator. This supports reproducibility.",
    )


class VirtualECU(AsyncScript, ABC):
    """Spawn a virtual ECU for testing purposes"""

    EPILOG = "https://fraunhofer-aisec.github.io/gallia/uds/virtual_ecu.html"

    def __init__(self, config: VirtualECUConfig):
        super().__init__(config)
        self.config: VirtualECUConfig = config

    @abstractmethod
    def _server(self) -> UDSServer: ...

    async def main(self) -> None:
        server = self._server()

        target: TargetURI = self.config.target
        transport: UDSServerTransport

        if sys.platform.startswith("linux"):
            from gallia.services.uds.server import (
                ISOTPUDSServerTransport,
                UnixUDSServerTransport,
            )

            match target.scheme:
                case TransportScheme.TCP:
                    transport = TCPUDSServerTransport(server, target)
                case TransportScheme.ISOTP:
                    transport = ISOTPUDSServerTransport(server, target)
                case TransportScheme.UNIX_LINES:
                    transport = UnixUDSServerTransport(server, target)
                case _:
                    assert False

        if sys.platform.startswith("win32"):
            match target.scheme:
                case TransportScheme.TCP:
                    transport = TCPUDSServerTransport(server, target)
                case _:
                    assert False

        try:
            await server.setup()
            await transport.run()
        finally:
            await server.teardown()


class RngVirtualECU(VirtualECU):
    CONFIG_TYPE = RngVirtualECUConfig
    SHORT_HELP = "Virtual ECU with randomized behavior"

    def __init__(self, config: RngVirtualECUConfig):
        super().__init__(config)
        self.config: RngVirtualECUConfig = config

    def _server(self) -> RandomUDSServer:
        return RandomUDSServer(self.config.seed, self.config, self.config)


class DbVirtualECU(VirtualECU):
    CONFIG_TYPE = DbVirtualECUConfig
    SHORT_HELP = "Virtual ECU which mimics the behavior of an ECU according to logs in the database"

    def __init__(self, config: DbVirtualECUConfig):
        super().__init__(config)
        self.config: DbVirtualECUConfig = config

    def _server(self) -> DBUDSServer:
        return DBUDSServer(self.config.path, self.config.ecu, self.config.properties, self.config)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class RMBAPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt = Field(0x01, description="The session in which the requests are made")
    address: AutoInt = Field(
        description="The start address from which data should be read", positional=True
    )
    length: AutoInt = Field(description="The number of bytes which should be read", positional=True)


class RMBAPrimitive(UDSScanner):
    """Read memory by address"""

    CONFIG_TYPE = RMBAPrimitiveConfig
    SHORT_HELP = "ReadMemoryByAddress"

    def __init__(self, config: RMBAPrimitiveConfig):
        super().__init__(config)
        self.config: RMBAPrimitiveConfig = config

    async def main(self) -> None:
        try:
            await self.ecu.check_and_set_session(self.config.session)
        except Exception as e:
            logger.critical(f"Could not change to session: {g_repr(self.config.session)}: {e!r}")
            sys.exit(1)

        resp = await self.ecu.read_memory_by_address(self.config.address, self.config.length)

        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            logger.result("Positive response:")

            logger.result(f"hex: {resp.data_record.hex()}")
            logger.result(f"raw: {repr(resp.data_record)}")



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC
from typing import Any, Literal, Self

from pydantic import field_serializer, model_validator

from gallia.cli.gallia import parse_and_run
from gallia.command import AsyncScript
from gallia.command.base import AsyncScriptConfig, ScannerConfig
from gallia.command.config import Field, Idempotent
from gallia.power_supply import power_supply_drivers
from gallia.power_supply.base import BasePowerSupplyDriver
from gallia.power_supply.uri import PowerSupplyURI
from gallia.transports import TargetURI
from gallia.utils import strtobool


class CLIConfig(AsyncScriptConfig):
    power_supply: Idempotent[PowerSupplyURI] = Field(
        description="URI specifying the location of the powersupply",
        metavar="URI",
        short="t",
        config_section=ScannerConfig._config_section,
    )
    channel: int = Field(description="the channel number to control", short="c")
    attr: Literal["voltage", "current", "output"] = Field(
        description="the attribute to control", short="a"
    )

    @field_serializer("power_supply")
    def serialize_target_uri(self, target_uri: TargetURI | None) -> Any:
        if target_uri is None:
            return None

        return target_uri.raw

    @model_validator(mode="after")
    def check_power_supply_requirements(self) -> Self:
        if self.power_supply.product_id not in power_supply_drivers:
            raise ValueError(f"powersupply {self.power_supply.product_id} is not supported")

        return self


class GetCLIConfig(CLIConfig):
    pass


class SetCLIConfig(CLIConfig):
    value: str = Field(positional=True)


class CLI(AsyncScript, ABC):
    def __init__(self, config: CLIConfig):
        super().__init__(config)
        self.config: CLIConfig = config

    async def _client(self) -> BasePowerSupplyDriver:
        driver = power_supply_drivers[self.config.power_supply.product_id]
        return await driver.connect(self.config.power_supply, timeout=1.0)


class GetCLI(CLI):
    CONFIG_TYPE = GetCLIConfig
    SHORT_HELP = "Get properties of the power supply"

    def __init__(self, config: GetCLIConfig):
        super().__init__(config)
        self.config: GetCLIConfig = config

    async def main(self) -> None:
        client = await self._client()

        match self.config.attr:
            case "voltage":
                print(await client.get_voltage(self.config.channel))
            case "current":
                print(await client.get_current(self.config.channel))
            case "output":
                if self.config.channel == 0:
                    print(await client.get_master())
                else:
                    print(await client.get_output(self.config.channel))


class SetCLI(CLI):
    CONFIG_TYPE = SetCLIConfig
    SHORT_HELP = "Set properties of the power supply"

    def __init__(self, config: SetCLIConfig):
        super().__init__(config)
        self.config: SetCLIConfig = config

    async def main(self) -> None:
        client = await self._client()

        match self.config.attr:
            case "voltage":
                await client.set_voltage(self.config.channel, float(self.config.value))
            case "current":
                await client.set_current(self.config.channel, float(self.config.value))
            case "output":
                if self.config.channel == 0:
                    await client.set_master(strtobool(self.config.value))
                else:
                    await client.set_output(self.config.channel, strtobool(self.config.value))


def main() -> None:
    parse_and_run({"get": GetCLI, "set": SetCLI})


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.service import NegativeResponse

logger = get_logger(__name__)


class ReadByIdentifierPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    data_identifier: AutoInt = Field(description="The data identifier", positional=True)
    session: AutoInt = Field(0x01, description="set session perform test in")


class ReadByIdentifierPrimitive(UDSScanner):
    """Read data via the ReadDataByIdentifier service"""

    CONFIG_TYPE = ReadByIdentifierPrimitiveConfig
    SHORT_HELP = "ReadDataByIdentifier"

    def __init__(self, config: ReadByIdentifierPrimitiveConfig):
        super().__init__(config)
        self.config: ReadByIdentifierPrimitiveConfig = config
        self.result: bytes | None = None

    async def main(self) -> None:
        try:
            if self.config.session != 0x01:
                await self.ecu.set_session(self.config.session)
        except Exception as e:
            logger.critical(f"fatal error: {e!r}")
            sys.exit(1)

        resp = await self.ecu.read_data_by_identifier(self.config.data_identifier)
        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            logger.result("Positive response:")
            data = resp.data_record
            logger.result(f"hex: {data.hex()}")
            logger.result(f"raw: {repr(data)}")
            self.result = data



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSRequest, UDSResponse
from gallia.services.uds.core.exception import UDSException
from gallia.services.uds.core.service import RawRequest, RawResponse
from gallia.services.uds.helpers import raise_for_error

logger = get_logger(__name__)


class SendPDUPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    pdus: list[HexBytes] = Field(description="The raw PDU(s) to send to the ECU", positional=True)
    session: AutoInt | None = Field(
        None, description="Change to this session prior to sending the PDU(s)"
    )


class SendPDUPrimitive(UDSScanner):
    """A raw scanner to send plain PDU(s)"""

    CONFIG_TYPE = SendPDUPrimitiveConfig
    SHORT_HELP = "send plain PDU(s)"

    def __init__(self, config: SendPDUPrimitiveConfig):
        super().__init__(config)
        self.config: SendPDUPrimitiveConfig = config

    async def _send_pdu(self, pdu: HexBytes) -> None:
        parsed_request = UDSRequest.parse_dynamic(pdu)

        if isinstance(parsed_request, RawRequest):
            logger.warning("Could not parse the request PDU")

        logger.result(f"Sending {parsed_request}")

        try:
            response = await self.ecu.send_raw(pdu)
        except UDSException as e:
            logger.error(repr(e))
            return

        if isinstance(response, NegativeResponse):
            logger.warning(f"Received {response}")
        else:
            if isinstance(response, RawResponse):
                logger.warning("Could not parse the response PDU")
            logger.result(f"Received {response}")

    async def main(self) -> None:
        if self.config.session is not None:
            resp: UDSResponse = await self.ecu.set_session(self.config.session)
            raise_for_error(resp)
            logger.result(f"Switched to session 0x{self.config.session:02x}")

        for pdu in self.config.pdus:
            await self._send_pdu(pdu)



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Declarative and Typed Argument Parsing.

This package contains the classes and methods required for declarative and
typed argument parsing.

The public interface exposed by this package is the `ArgumentParser` class,
which is intended to be a *near* drop-in replacement for the Python standard
library `argparse.ArgumentParser` - while providing declarative and typed
argument parsing.
"""

# Local
from gallia.pydantic_argparse.argparse.parser import ArgumentParser

# Public Re-Exports
__all__ = ("ArgumentParser",)
